/*
 * Copyright (C) 2013-2014 Red Hat, Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Peter Schiffer <pschiffe@redhat.com>
 */

#include <time.h>
#include <limits.h>
#include <konkret/konkret.h>
#include "LMI_BIOSElement.h"
#include "utils.h"
#include "dmidecode.h"

static const CMPIBroker* _cb = NULL;

static void LMI_BIOSElementInitialize(const CMPIContext *ctx)
{
    lmi_init(provider_name, _cb, ctx, provider_config_defaults);
}

static CMPIStatus LMI_BIOSElementCleanup(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_BIOSElementEnumInstanceNames(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_BIOSElementEnumInstances(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    LMI_BIOSElement lmi_bios;
    const char *ns = KNameSpace(cop);
    char id[BUFLEN], instance_id[BUFLEN], *dot;
    DmiBios dmi_bios;
    unsigned i, number = 0;
    struct tm tm;

    if (dmi_get_bios(&dmi_bios) != 0) {
        goto done;
    }

    snprintf(id, BUFLEN, "%s BIOS %s", dmi_bios.vendor, dmi_bios.version);
    snprintf(instance_id, BUFLEN,
            LMI_ORGID ":" LMI_BIOSElement_ClassName ":%s", id);

    LMI_BIOSElement_Init(&lmi_bios, _cb, ns);

    LMI_BIOSElement_Set_SoftwareElementState(&lmi_bios,
            LMI_BIOSElement_SoftwareElementState_Running);
    LMI_BIOSElement_Set_TargetOperatingSystem(&lmi_bios,
            LMI_BIOSElement_TargetOperatingSystem_Not_Applicable);
    LMI_BIOSElement_Set_PrimaryBIOS(&lmi_bios, 1);
    LMI_BIOSElement_Set_Caption(&lmi_bios, "BIOS");
    LMI_BIOSElement_Set_Description(&lmi_bios,
            "This object represents BIOS of the system.");

    LMI_BIOSElement_Set_Version(&lmi_bios, dmi_bios.version);
    LMI_BIOSElement_Set_Name(&lmi_bios, id);
    LMI_BIOSElement_Set_ElementName(&lmi_bios, id);
    LMI_BIOSElement_Set_SoftwareElementID(&lmi_bios, id);
    LMI_BIOSElement_Set_InstanceID(&lmi_bios, instance_id);
    LMI_BIOSElement_Set_Manufacturer(&lmi_bios, dmi_bios.vendor);
    LMI_BIOSElement_Set_LoadedStartingAddress(&lmi_bios, dmi_bios.address);
    LMI_BIOSElement_Set_CurrentLanguage(&lmi_bios, dmi_bios.curr_lang);

    if (*dmi_bios.bios_revision) {
        sscanf(dmi_bios.bios_revision, "%u", &number);
        if (number < UCHAR_MAX) {
            LMI_BIOSElement_Set_SystemBIOSMajorRelease(&lmi_bios, number);
        }
        number = 0;
    }
    if ((dot = strchr(dmi_bios.bios_revision, '.'))) {
        sscanf(dot + 1, "%u", &number);
        if (number < UCHAR_MAX) {
            LMI_BIOSElement_Set_SystemBIOSMinorRelease(&lmi_bios, number);
        }
        number = 0;
    }
    if (*dmi_bios.fw_revision) {
        sscanf(dmi_bios.fw_revision, "%u", &number);
        if (number < UCHAR_MAX) {
            LMI_BIOSElement_Set_EmbeddedControllerFirmwareMajorRelease(
                    &lmi_bios, number);
        }
        number = 0;
    }
    if ((dot = strchr(dmi_bios.fw_revision, '.'))) {
        sscanf(dot + 1, "%u", &number);
        if (number < UCHAR_MAX) {
            LMI_BIOSElement_Set_EmbeddedControllerFirmwareMinorRelease(
                    &lmi_bios, number);
        }
        number = 0;
    }

    memset(&tm, 0, sizeof(struct tm));
    if (strptime(dmi_bios.release_date, "%m/%d/%Y", &tm)) {
        LMI_BIOSElement_Set_ReleaseDate(&lmi_bios,
                CMNewDateTimeFromBinary(_cb, mktime(&tm) * 1000000, 0, NULL));
    }

    if (dmi_bios.langs_nb > 0) {
        LMI_BIOSElement_Init_ListOfLanguages(&lmi_bios, dmi_bios.langs_nb);
        for (i = 0; i < dmi_bios.langs_nb; i++) {
            LMI_BIOSElement_Set_ListOfLanguages(&lmi_bios, i, dmi_bios.langs[i]);
        }
    }

    if (dmi_bios.langs_nb > 1) {
        LMI_BIOSElement_Set_LanguageEdition(&lmi_bios, "Multilingual");
    } else {
        LMI_BIOSElement_Set_LanguageEdition(&lmi_bios, dmi_bios.curr_lang);
    }

    KReturnInstance(cr, lmi_bios);

done:
    dmi_free_bios(&dmi_bios);

    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_BIOSElementGetInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_BIOSElementCreateInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_BIOSElementModifyInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci,
    const char** properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_BIOSElementDeleteInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_BIOSElementExecQuery(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* lang,
    const char* query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

CMInstanceMIStub(
    LMI_BIOSElement,
    LMI_BIOSElement,
    _cb,
    LMI_BIOSElementInitialize(ctx))

static CMPIStatus LMI_BIOSElementMethodCleanup(
    CMPIMethodMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_BIOSElementInvokeMethod(
    CMPIMethodMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* meth,
    const CMPIArgs* in,
    CMPIArgs* out)
{
    return LMI_BIOSElement_DispatchMethod(
        _cb, mi, cc, cr, cop, meth, in, out);
}

CMMethodMIStub(
    LMI_BIOSElement,
    LMI_BIOSElement,
    _cb,
    LMI_BIOSElementInitialize(ctx))

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_BIOSElement",
    "LMI_BIOSElement",
    "instance method")
