/*
 * Copyright (C) 2013-2014 Red Hat, Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Peter Schiffer <pschiffe@redhat.com>
 */

#include <konkret/konkret.h>
#include "LMI_BIOSFeature.h"
#include "utils.h"
#include "dmidecode.h"

CMPIUint16 get_bios_characteristic(const char *dmi_charact);

static const CMPIBroker* _cb = NULL;

static void LMI_BIOSFeatureInitialize(const CMPIContext *ctx)
{
    lmi_init(provider_name, _cb, ctx, provider_config_defaults);
}

static CMPIStatus LMI_BIOSFeatureCleanup(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_BIOSFeatureEnumInstanceNames(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_BIOSFeatureEnumInstances(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    LMI_BIOSFeature lmi_bios;
    const char *ns = KNameSpace(cop);
    char id[BUFLEN], instance_id[BUFLEN];
    DmiBios dmi_bios;
    unsigned i;

    if (dmi_get_bios(&dmi_bios) != 0) {
        goto done;
    }

    snprintf(id, BUFLEN, "%s BIOS %s", dmi_bios.vendor, dmi_bios.version);
    snprintf(instance_id, BUFLEN,
            LMI_ORGID ":" LMI_BIOSFeature_ClassName ":%s", id);

    LMI_BIOSFeature_Init(&lmi_bios, _cb, ns);

    LMI_BIOSFeature_Set_Name(&lmi_bios, id);
    LMI_BIOSFeature_Set_ElementName(&lmi_bios, id);
    LMI_BIOSFeature_Set_ProductName(&lmi_bios, id);
    LMI_BIOSFeature_Set_Vendor(&lmi_bios, dmi_bios.vendor);
    LMI_BIOSFeature_Set_Version(&lmi_bios, dmi_bios.version);
    LMI_BIOSFeature_Set_IdentifyingNumber(&lmi_bios, dmi_bios.version);
    LMI_BIOSFeature_Set_InstanceID(&lmi_bios, instance_id);
    LMI_BIOSFeature_Set_Caption(&lmi_bios, "BIOS Feature");
    LMI_BIOSFeature_Set_Description(&lmi_bios,
            "This object represents BIOS Features of the system.");

    if (dmi_bios.charact_nb) {
        LMI_BIOSFeature_Init_Characteristics(&lmi_bios, dmi_bios.charact_nb);
        LMI_BIOSFeature_Init_CharacteristicDescriptions(&lmi_bios, dmi_bios.charact_nb);
        for (i = 0; i < dmi_bios.charact_nb; i++) {
            LMI_BIOSFeature_Set_Characteristics(&lmi_bios, i,
                    get_bios_characteristic(dmi_bios.characteristics[i]));
            LMI_BIOSFeature_Set_CharacteristicDescriptions(&lmi_bios, i,
                    dmi_bios.characteristics[i]);
        }
    }

    KReturnInstance(cr, lmi_bios);

done:
    dmi_free_bios(&dmi_bios);

    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_BIOSFeatureGetInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_BIOSFeatureCreateInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_BIOSFeatureModifyInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci,
    const char** properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_BIOSFeatureDeleteInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_BIOSFeatureExecQuery(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* lang,
    const char* query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

CMInstanceMIStub(
    LMI_BIOSFeature,
    LMI_BIOSFeature,
    _cb,
    LMI_BIOSFeatureInitialize(ctx))

static CMPIStatus LMI_BIOSFeatureMethodCleanup(
    CMPIMethodMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_BIOSFeatureInvokeMethod(
    CMPIMethodMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* meth,
    const CMPIArgs* in,
    CMPIArgs* out)
{
    return LMI_BIOSFeature_DispatchMethod(
        _cb, mi, cc, cr, cop, meth, in, out);
}

/*
 * Get BIOS characteristics according to the dmidecode output.
 * @param dmi_charact characteristics from dmidecode
 * @return characteristic defined by CIM
 */
CMPIUint16 get_bios_characteristic(const char *dmi_charact)
{
    if (!dmi_charact) {
        return 2; /* Unknown */
    }

    /* CIM characteristics based on
       cim_schema_2.42.0Final-MOFs/Application/CIM_BIOSFeature.mof */
    /* Dmidecode characts based on dmidecode 2.12 */
    static struct {
        CMPIUint16 value_map;   /* ValueMap defined by CIM */
        char *value;            /* Value defined by CIM (just for reference) */
        char *search;           /* String used to match the dmidecode charact */
    } characts[] = {
        /*
        {1, "Other", ""},
        {2, "Unknown", ""},
        {3, "Undefined", ""},
        */
        {4, "ISA Support", "ISA is supported"},
        {5, "MCA Support", "MCA is supported"},
        {6, "EISA Support", "EISA is supported"},
        {7, "PCI Support", "PCI is supported"},
        {8, "PCMCIA Support", "PC Card (PCMCIA) is supported"},
        {9, "PnP Support", "PNP is supported"},
        {10, "APM Support", "APM is supported"},
        {11, "Upgradeable BIOS", "BIOS is upgradeable"},
        {12, "BIOS Shadowing Allowed", "BIOS shadowing is allowed"},
        /*
        {13, "VL VESA Support", ""},
        */
        {14, "ESCD Support", "ESCD support is available"},
        {15, "LS-120 Boot Support", "LS-120 boot is supported"},
        {16, "ACPI Support", "ACPI is supported"},
        {17, "I2O Boot Support", "I2O boot is supported"},
        {18, "USB Legacy Support", "USB legacy is supported"},
        {19, "AGP Support", "AGP is supported"},
        /*
        {20, "PC Card", ""},
        {21, "IR", ""},
        {22, "1394", ""},
        {23, "I2C", ""},
        */
        {24, "Smart Battery", "Smart battery is supported"},
        {25, "ATAPI ZIP Drive Boot Support", "ATAPI Zip drive boot is supported"},
        {26, "1394 Boot Support", "IEEE 1394 boot is supported"},
        {27, "Boot from CD", "Boot from CD is supported"},
        {28, "Selectable Boot", "Selectable boot is supported"},
        {29, "BIOS ROM is Socketed", "BIOS ROM is socketed"},
        {30, "Boot from PCMCIA", "Boot from PC Card (PCMCIA) is supported"},
        {31, "EDD Specification Support", "EDD is supported"},
        {160, "PC-98", "NEC PC-98"},
    };

    size_t i, characts_length = sizeof(characts) / sizeof(characts[0]);

    for (i = 0; i < characts_length; i++) {
        if (strcmp(dmi_charact, characts[i].search) == 0) {
            return characts[i].value_map;
        }
    }

    return 1; /* Other */
}

CMMethodMIStub(
    LMI_BIOSFeature,
    LMI_BIOSFeature,
    _cb,
    LMI_BIOSFeatureInitialize(ctx))

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_BIOSFeature",
    "LMI_BIOSFeature",
    "instance method")
