/*
 * Copyright (C) 2013-2014 Red Hat, Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Peter Schiffer <pschiffe@redhat.com>
 */

#include <konkret/konkret.h>
#include "LMI_PCIControlledBy.h"
#include "LMI_PCIDevice.h"
#include "LMI_PCIBridge.h"
#include "utils.h"
#include "PCIDev.h"

static const CMPIBroker* _cb;

struct pci_access *acc_ctrlby = NULL;

static void LMI_PCIControlledByInitialize(const CMPIContext *ctx)
{
    lmi_init(provider_name, _cb, ctx, provider_config_defaults);

    if (init_pci_access(&acc_ctrlby, PCI_FILL_IDENT
                                | PCI_FILL_IRQ
                                | PCI_FILL_BASES
                                | PCI_FILL_ROM_BASE
                                | PCI_FILL_CLASS
                                | PCI_FILL_CAPS) != 0) {
        lmi_error("Failed to access the PCI bus.");
        abort();
    }
}

static CMPIStatus LMI_PCIControlledByCleanup(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    cleanup_pci_access(&acc_ctrlby);

    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIControlledByEnumInstanceNames(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_PCIControlledByEnumInstances(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    LMI_PCIControlledBy lmi_assoc;
    LMI_PCIDeviceRef lmi_dev;
    LMI_PCIBridgeRef lmi_brg;
    LMI_PCIPortRef lmi_port;
    const char *ns = KNameSpace(cop);
    struct pci_dev *dev;
    u8 secondary_bus;
    char device_id_str[BUFLEN], port_id[BUFLEN];
    unsigned short i, ports;

    for (dev = acc_ctrlby->devices; dev; dev = dev->next) {
        secondary_bus = pci_read_byte(dev, PCI_SECONDARY_BUS);
        /* Bridges are connected to two PCI buses. */
        /* Throw away the lower 8 bits denoting the subclass */
        if (((dev->device_class) >> 8) == LMI_PCIDevice_ClassCode_Bridge && secondary_bus) {
            ports = 2;
        } else {
            ports = 1;
        }

        snprintf(device_id_str, BUFLEN, "%02x:%02x.%u", dev->bus, dev->dev,
                dev->func);

        /* Create correct instance for PCIDevice and PCIBridge */
        /* Throw away the lower 8 bits denoting the subclass */
        if (((dev->device_class) >> 8) == LMI_PCIDevice_ClassCode_Bridge) {
            /* PCIBridge */
            LMI_PCIBridgeRef_Init(&lmi_brg, _cb, ns);
            LMI_PCIBridgeRef_Set_SystemCreationClassName(&lmi_brg,
                    lmi_get_system_creation_class_name());
            LMI_PCIBridgeRef_Set_SystemName(&lmi_brg, lmi_get_system_name_safe(cc));
            LMI_PCIBridgeRef_Set_CreationClassName(&lmi_brg, LMI_PCIBridge_ClassName);
            LMI_PCIBridgeRef_Set_DeviceID(&lmi_brg, device_id_str);
        } else {
            /* PCIDevice */
            LMI_PCIDeviceRef_Init(&lmi_dev, _cb, ns);
            LMI_PCIDeviceRef_Set_SystemCreationClassName(&lmi_dev,
                    lmi_get_system_creation_class_name());
            LMI_PCIDeviceRef_Set_SystemName(&lmi_dev, lmi_get_system_name_safe(cc));
            LMI_PCIDeviceRef_Set_CreationClassName(&lmi_dev, LMI_PCIDevice_ClassName);
            LMI_PCIDeviceRef_Set_DeviceID(&lmi_dev, device_id_str);
        }

        for (i = 0; i < ports; i++) {
            if (i == 0) {
                snprintf(port_id, BUFLEN, "%02x:%02x.%u-%x", dev->bus, dev->dev,
                        dev->func, dev->bus);
            } else {
                snprintf(port_id, BUFLEN, "%02x:%02x.%u-%x", dev->bus, dev->dev,
                        dev->func, secondary_bus);
            }

            LMI_PCIPortRef_Init(&lmi_port, _cb, ns);
            LMI_PCIPortRef_Set_SystemCreationClassName(&lmi_port,
                    lmi_get_system_creation_class_name());
            LMI_PCIPortRef_Set_SystemName(&lmi_port, lmi_get_system_name_safe(cc));
            LMI_PCIPortRef_Set_CreationClassName(&lmi_port, LMI_PCIPort_ClassName);
            LMI_PCIPortRef_Set_DeviceID(&lmi_port, port_id);

            LMI_PCIControlledBy_Init(&lmi_assoc, _cb, ns);
            LMI_PCIControlledBy_Set_Dependent(&lmi_assoc, &lmi_port);

            /* Choose correct instance for PCIDevice and PCIBridge */
            /* Throw away the lower 8 bits denoting the subclass */
            if (((dev->device_class) >> 8) == LMI_PCIDevice_ClassCode_Bridge) {
                LMI_PCIControlledBy_Set_Antecedent(&lmi_assoc,
                        (CIM_PCIDeviceRef *) &lmi_brg);
            } else {
                LMI_PCIControlledBy_Set_Antecedent(&lmi_assoc,
                        (CIM_PCIDeviceRef *) &lmi_dev);
            }

            KReturnInstance(cr, lmi_assoc);
        }
    }

    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIControlledByGetInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_PCIControlledByCreateInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIControlledByModifyInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci,
    const char**properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIControlledByDeleteInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIControlledByExecQuery(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* lang,
    const char* query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIControlledByAssociationCleanup(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIControlledByAssociators(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* resultClass,
    const char* role,
    const char* resultRole,
    const char** properties)
{
    return KDefaultAssociators(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_PCIControlledBy_ClassName,
        assocClass,
        resultClass,
        role,
        resultRole,
        properties);
}

static CMPIStatus LMI_PCIControlledByAssociatorNames(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* resultClass,
    const char* role,
    const char* resultRole)
{
    return KDefaultAssociatorNames(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_PCIControlledBy_ClassName,
        assocClass,
        resultClass,
        role,
        resultRole);
}

static CMPIStatus LMI_PCIControlledByReferences(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* role,
    const char** properties)
{
    return KDefaultReferences(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_PCIControlledBy_ClassName,
        assocClass,
        role,
        properties);
}

static CMPIStatus LMI_PCIControlledByReferenceNames(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* role)
{
    return KDefaultReferenceNames(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_PCIControlledBy_ClassName,
        assocClass,
        role);
}

CMInstanceMIStub(
    LMI_PCIControlledBy,
    LMI_PCIControlledBy,
    _cb,
    LMI_PCIControlledByInitialize(ctx))

CMAssociationMIStub(
    LMI_PCIControlledBy,
    LMI_PCIControlledBy,
    _cb,
    LMI_PCIControlledByInitialize(ctx))

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_PCIControlledBy",
    "LMI_PCIControlledBy",
    "instance association")
