/*
 * Copyright (C) 2013-2014 Red Hat, Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Peter Schiffer <pschiffe@redhat.com>
 */

#include <konkret/konkret.h>
#include "LMI_PCIPortGroup.h"
#include "LMI_PCIDevice.h"
#include "utils.h"
#include "PCIDev.h"

static const CMPIBroker* _cb = NULL;

struct pci_access *acc_portgroups = NULL;

static void LMI_PCIPortGroupInitialize(const CMPIContext *ctx)
{
    lmi_init(provider_name, _cb, ctx, provider_config_defaults);

    if (init_pci_access(&acc_portgroups, PCI_FILL_IDENT
                                | PCI_FILL_IRQ
                                | PCI_FILL_BASES
                                | PCI_FILL_ROM_BASE
                                | PCI_FILL_CLASS
                                | PCI_FILL_CAPS) != 0) {
        lmi_error("Failed to access the PCI bus.");
        abort();
    }
}

static CMPIStatus LMI_PCIPortGroupCleanup(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    cleanup_pci_access(&acc_portgroups);

    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIPortGroupEnumInstanceNames(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_PCIPortGroupEnumInstances(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    LMI_PCIPortGroup lmi_dev;
    const char *ns = KNameSpace(cop);
    struct pci_dev *dev;
    u8 secondary_bus, buses[BUFLEN] = { 0 };
    char instance_id[BUFLEN], name[BUFLEN];
    short i, buses_nb = 0;

    for (dev = acc_portgroups->devices; dev; dev = dev->next) {
        add_bus(buses, &buses_nb, BUFLEN, dev->bus);

        secondary_bus = pci_read_byte(dev, PCI_SECONDARY_BUS);
        /* Bridges are connected to two PCI buses. */
        /* Throw away the lower 8 bits denoting the subclass */
        if (((dev->device_class) >> 8) == LMI_PCIDevice_ClassCode_Bridge && secondary_bus) {
            add_bus(buses, &buses_nb, BUFLEN, secondary_bus);
        }
    }

    for (i = 0; i < buses_nb; i++) {
        snprintf(instance_id, BUFLEN,
                LMI_ORGID ":" LMI_PCIPortGroup_ClassName ":%d", buses[i]);
        snprintf(name, BUFLEN, "PCI Bus Number %d", buses[i]);

        LMI_PCIPortGroup_Init(&lmi_dev, _cb, ns);

        LMI_PCIPortGroup_Set_InstanceID(&lmi_dev, instance_id);
        LMI_PCIPortGroup_Set_BusNumber(&lmi_dev, buses[i]);
        LMI_PCIPortGroup_Set_Caption(&lmi_dev, "PCI Port Group");
        LMI_PCIPortGroup_Set_Description(&lmi_dev,
                "This object represents one PCI bus contained in system.");
        LMI_PCIPortGroup_Set_Name(&lmi_dev, name);
        LMI_PCIPortGroup_Set_ElementName(&lmi_dev, name);

        KReturnInstance(cr, lmi_dev);
    }

    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIPortGroupGetInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_PCIPortGroupCreateInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIPortGroupModifyInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci,
    const char** properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIPortGroupDeleteInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIPortGroupExecQuery(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* lang,
    const char* query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

CMInstanceMIStub(
    LMI_PCIPortGroup,
    LMI_PCIPortGroup,
    _cb,
    LMI_PCIPortGroupInitialize(ctx))

static CMPIStatus LMI_PCIPortGroupMethodCleanup(
    CMPIMethodMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIPortGroupInvokeMethod(
    CMPIMethodMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* meth,
    const CMPIArgs* in,
    CMPIArgs* out)
{
    return LMI_PCIPortGroup_DispatchMethod(
        _cb, mi, cc, cr, cop, meth, in, out);
}

CMMethodMIStub(
    LMI_PCIPortGroup,
    LMI_PCIPortGroup,
    _cb,
    LMI_PCIPortGroupInitialize(ctx))

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_PCIPortGroup",
    "LMI_PCIPortGroup",
    "instance method")
