/*
 * Copyright (C) 2013-2014 Red Hat, Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Peter Schiffer <pschiffe@redhat.com>
 */

#include <konkret/konkret.h>
#include "LMI_PCIPortMemberOfCollection.h"
#include "LMI_PCIDevice.h"
#include "utils.h"
#include "PCIDev.h"

static const CMPIBroker* _cb;

struct pci_access *acc_memberofcoll = NULL;

static void LMI_PCIPortMemberOfCollectionInitialize(const CMPIContext *ctx)
{
    lmi_init(provider_name, _cb, ctx, provider_config_defaults);

    if (init_pci_access(&acc_memberofcoll, PCI_FILL_IDENT
                                | PCI_FILL_IRQ
                                | PCI_FILL_BASES
                                | PCI_FILL_ROM_BASE
                                | PCI_FILL_CLASS
                                | PCI_FILL_CAPS) != 0) {
        lmi_error("Failed to access the PCI bus.");
        abort();
    }
}

static CMPIStatus LMI_PCIPortMemberOfCollectionCleanup(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    cleanup_pci_access(&acc_memberofcoll);

    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionEnumInstanceNames(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionEnumInstances(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    LMI_PCIPortMemberOfCollection lmi_assoc;
    LMI_PCIPortGroupRef lmi_dev;
    LMI_PCIPortRef lmi_port;
    const char *ns = KNameSpace(cop);
    struct pci_dev *dev;
    u8 secondary_bus;
    char port_id[BUFLEN], instance_id[BUFLEN];
    unsigned short i, ports;

    for (dev = acc_memberofcoll->devices; dev; dev = dev->next) {
        secondary_bus = pci_read_byte(dev, PCI_SECONDARY_BUS);
        /* Bridges are connected to two PCI buses. */
        /* Throw away the lower 8 bits denoting the subclass */
        if (((dev->device_class) >> 8) == LMI_PCIDevice_ClassCode_Bridge && secondary_bus) {
            ports = 2;
        } else {
            ports = 1;
        }

        for (i = 0; i < ports; i++) {
            if (i == 0) {
                snprintf(port_id, BUFLEN, "%02x:%02x.%u-%x", dev->bus, dev->dev,
                        dev->func, dev->bus);
            } else {
                snprintf(port_id, BUFLEN, "%02x:%02x.%u-%x", dev->bus, dev->dev,
                        dev->func, secondary_bus);
            }

            LMI_PCIPortRef_Init(&lmi_port, _cb, ns);
            LMI_PCIPortRef_Set_SystemCreationClassName(&lmi_port,
                    lmi_get_system_creation_class_name());
            LMI_PCIPortRef_Set_SystemName(&lmi_port, lmi_get_system_name_safe(cc));
            LMI_PCIPortRef_Set_CreationClassName(&lmi_port, LMI_PCIPort_ClassName);
            LMI_PCIPortRef_Set_DeviceID(&lmi_port, port_id);

            snprintf(instance_id, BUFLEN,
                    LMI_ORGID ":" LMI_PCIPortGroup_ClassName ":%d", dev->bus);

            LMI_PCIPortGroupRef_Init(&lmi_dev, _cb, ns);
            LMI_PCIPortGroupRef_Set_InstanceID(&lmi_dev, instance_id);

            LMI_PCIPortMemberOfCollection_Init(&lmi_assoc, _cb, ns);
            LMI_PCIPortMemberOfCollection_Set_Member(&lmi_assoc, &lmi_port);
            LMI_PCIPortMemberOfCollection_Set_Collection(&lmi_assoc, &lmi_dev);

            KReturnInstance(cr, lmi_assoc);
        }
    }

    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionGetInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionCreateInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionModifyInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci,
    const char**properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionDeleteInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionExecQuery(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* lang,
    const char* query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionAssociationCleanup(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionAssociators(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* resultClass,
    const char* role,
    const char* resultRole,
    const char** properties)
{
    return KDefaultAssociators(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_PCIPortMemberOfCollection_ClassName,
        assocClass,
        resultClass,
        role,
        resultRole,
        properties);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionAssociatorNames(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* resultClass,
    const char* role,
    const char* resultRole)
{
    return KDefaultAssociatorNames(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_PCIPortMemberOfCollection_ClassName,
        assocClass,
        resultClass,
        role,
        resultRole);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionReferences(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* role,
    const char** properties)
{
    return KDefaultReferences(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_PCIPortMemberOfCollection_ClassName,
        assocClass,
        role,
        properties);
}

static CMPIStatus LMI_PCIPortMemberOfCollectionReferenceNames(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* role)
{
    return KDefaultReferenceNames(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_PCIPortMemberOfCollection_ClassName,
        assocClass,
        role);
}

CMInstanceMIStub(
    LMI_PCIPortMemberOfCollection,
    LMI_PCIPortMemberOfCollection,
    _cb,
    LMI_PCIPortMemberOfCollectionInitialize(ctx))

CMAssociationMIStub(
    LMI_PCIPortMemberOfCollection,
    LMI_PCIPortMemberOfCollection,
    _cb,
    LMI_PCIPortMemberOfCollectionInitialize(ctx))

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_PCIPortMemberOfCollection",
    "LMI_PCIPortMemberOfCollection",
    "instance association")
