/*
 * Copyright (C) 2013-2014 Red Hat, Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Peter Schiffer <pschiffe@redhat.com>
 */

#include <konkret/konkret.h>
#include "LMI_PCIDevice.h"
#include "LMI_PCIPort.h"
#include "utils.h"
#include "PCIDev.h"

static const CMPIBroker* _cb = NULL;

struct pci_access *acc_port = NULL;

static void LMI_PCIPortInitialize(const CMPIContext *ctx)
{
    lmi_init(provider_name, _cb, ctx, provider_config_defaults);

    if (init_pci_access(&acc_port, PCI_FILL_IDENT
                                | PCI_FILL_IRQ
                                | PCI_FILL_BASES
                                | PCI_FILL_ROM_BASE
                                | PCI_FILL_CLASS
                                | PCI_FILL_CAPS) != 0) {
        lmi_error("Failed to access the PCI bus.");
        abort();
    }
}

static CMPIStatus LMI_PCIPortCleanup(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    cleanup_pci_access(&acc_port);

    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIPortEnumInstanceNames(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_PCIPortEnumInstances(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    LMI_PCIPort lmi_dev;
    const char *ns = KNameSpace(cop);
    struct pci_dev *dev;
    u8 secondary_bus;
    char port_id[BUFLEN], name[BUFLEN], instance_id[BUFLEN];
    unsigned short i, ports, port_type;

    for (dev = acc_port->devices; dev; dev = dev->next) {
        secondary_bus = pci_read_byte(dev, PCI_SECONDARY_BUS);
        if (((dev->device_class) >> 8) == LMI_PCIDevice_ClassCode_Bridge && secondary_bus) {
            ports = 2;
        } else {
            ports = 1;
        }

        for (i = 0; i < ports; i++) {
            LMI_PCIPort_Init(&lmi_dev, _cb, ns);

            LMI_PCIPort_Set_SystemCreationClassName(&lmi_dev,
                    lmi_get_system_creation_class_name());
            LMI_PCIPort_Set_SystemName(&lmi_dev, lmi_get_system_name_safe(cc));
            LMI_PCIPort_Set_CreationClassName(&lmi_dev, LMI_PCIPort_ClassName);

            LMI_PCIPort_Set_Caption(&lmi_dev, "PCI Port");
            LMI_PCIPort_Set_Description(&lmi_dev,
                    "This object represents one PCI Port on PCI Device.");
            LMI_PCIPort_Set_EnabledState(&lmi_dev,
                    LMI_PCIPort_EnabledState_Enabled);

            if (i == 0) {
                snprintf(port_id, BUFLEN, "%02x:%02x.%u-%x", dev->bus, dev->dev,
                        dev->func, dev->bus);
            } else {
                snprintf(port_id, BUFLEN, "%02x:%02x.%u-%x", dev->bus, dev->dev,
                        dev->func, secondary_bus);
            }

            if (pci_find_cap(dev, PCI_CAP_ID_EXP, PCI_CAP_NORMAL)) {
                port_type = LMI_PCIPort_PortType_PCIe;
            } else if (pci_find_cap(dev, PCI_CAP_ID_PCIX, PCI_CAP_NORMAL)) {
                port_type = LMI_PCIPort_PortType_PCI_X;
            } else {
                port_type = LMI_PCIPort_PortType_PCI;
            }

            LMI_PCIPort_Set_DeviceID(&lmi_dev, port_id);
            LMI_PCIPort_Set_PortType(&lmi_dev, port_type);

            port_id[7] = '\0';
            snprintf(name, BUFLEN, "PCI Port on %s", port_id);
            port_id[7] = '-';
            LMI_PCIPort_Set_Name(&lmi_dev, name);
            LMI_PCIPort_Set_ElementName(&lmi_dev, name);

            snprintf(instance_id,
                    BUFLEN, LMI_ORGID ":" LMI_PCIPort_ClassName ":%s", port_id);
            LMI_PCIPort_Set_InstanceID(&lmi_dev, instance_id);

            KReturnInstance(cr, lmi_dev);
        }
    }

    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIPortGetInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_PCIPortCreateInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIPortModifyInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci,
    const char** properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIPortDeleteInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_PCIPortExecQuery(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* lang,
    const char* query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

CMInstanceMIStub(
    LMI_PCIPort,
    LMI_PCIPort,
    _cb,
    LMI_PCIPortInitialize(ctx))

static CMPIStatus LMI_PCIPortMethodCleanup(
    CMPIMethodMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_PCIPortInvokeMethod(
    CMPIMethodMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* meth,
    const CMPIArgs* in,
    CMPIArgs* out)
{
    return LMI_PCIPort_DispatchMethod(
        _cb, mi, cc, cr, cop, meth, in, out);
}

CMMethodMIStub(
    LMI_PCIPort,
    LMI_PCIPort,
    _cb,
    LMI_PCIPortInitialize(ctx))

KUint32 LMI_PCIPort_RequestStateChange(
    const CMPIBroker* cb,
    CMPIMethodMI* mi,
    const CMPIContext* context,
    const LMI_PCIPortRef* self,
    const KUint16* RequestedState,
    KRef* Job,
    const KDateTime* TimeoutPeriod,
    CMPIStatus* status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_PCIPort_SetPowerState(
    const CMPIBroker* cb,
    CMPIMethodMI* mi,
    const CMPIContext* context,
    const LMI_PCIPortRef* self,
    const KUint16* PowerState,
    const KDateTime* Time,
    CMPIStatus* status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_PCIPort_Reset(
    const CMPIBroker* cb,
    CMPIMethodMI* mi,
    const CMPIContext* context,
    const LMI_PCIPortRef* self,
    CMPIStatus* status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_PCIPort_EnableDevice(
    const CMPIBroker* cb,
    CMPIMethodMI* mi,
    const CMPIContext* context,
    const LMI_PCIPortRef* self,
    const KBoolean* Enabled,
    CMPIStatus* status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_PCIPort_OnlineDevice(
    const CMPIBroker* cb,
    CMPIMethodMI* mi,
    const CMPIContext* context,
    const LMI_PCIPortRef* self,
    const KBoolean* Online,
    CMPIStatus* status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_PCIPort_QuiesceDevice(
    const CMPIBroker* cb,
    CMPIMethodMI* mi,
    const CMPIContext* context,
    const LMI_PCIPortRef* self,
    const KBoolean* Quiesce,
    CMPIStatus* status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_PCIPort_SaveProperties(
    const CMPIBroker* cb,
    CMPIMethodMI* mi,
    const CMPIContext* context,
    const LMI_PCIPortRef* self,
    CMPIStatus* status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_PCIPort_RestoreProperties(
    const CMPIBroker* cb,
    CMPIMethodMI* mi,
    const CMPIContext* context,
    const LMI_PCIPortRef* self,
    CMPIStatus* status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_PCIPort",
    "LMI_PCIPort",
    "instance method")
