/*
 * Copyright (C) 2014 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Jan Synacek <jsynacek@redhat.com>
 *
 */

#include <konkret/konkret.h>
#include "LMI_AffectedSELinuxJobElement.h"
#include "LMI_SELinuxElement.h"
#include "LMI_UnixFile.h"
#include "selinux.h"
#include "job_manager.h"

static const CMPIBroker *_cb;

static void LMI_AffectedSELinuxJobElementInitialize(const CMPIContext *ctx)
{
    selinux_provider_init(LMI_SELinuxJob_ClassName, FALSE, provider_config_defaults, _cb, ctx);
}

static CMPIStatus LMI_AffectedSELinuxJobElementCleanup(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_AffectedSELinuxJobElementEnumInstanceNames(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_AffectedSELinuxJobElementEnumInstances(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    CMPIStatus st = {CMPI_RC_OK, NULL};
    LmiJob *job;
    gchar *jobid;
    guint *jobnum;
    guint *jobnums = jobmgr_get_job_numbers(NULL);
    CMPIObjectPath *o;
    const char *ns = KNameSpace(cop);
    const gchar *instance_id;
    LMI_AffectedSELinuxJobElement lmi_affected;
    gchar _type;

    if (jobnums == NULL)
        CMReturn(CMPI_RC_ERR_FAILED);

    for (jobnum = jobnums; *jobnum != 0; jobnum++) {
        job = jobmgr_get_job_by_number(*jobnum);
        if (!job)
            /* job may have been deleted since numbers query */
            continue;

        if (!LMI_IS_JOB(job)) {
            g_object_unref(job);
            continue;
        }

        st = jobmgr_job_to_cim_op(job, &o);
        if (st.rc != CMPI_RC_OK) {
            jobid = lmi_job_get_jobid(job);
            if (st.msg) {
                lmi_error("Failed to make cim instance out of job \"%s\": %s",
                               jobid, CMGetCharsPtr(st.msg, NULL));
            } else {
                lmi_error("Failed to make cim instance out of job \"%s\"", jobid);
            }
            g_free(jobid);
            continue;
        }
        g_object_unref(job);

        LMI_AffectedSELinuxJobElement_Init(&lmi_affected, _cb, ns);
        LMI_AffectedSELinuxJobElement_SetObjectPath_AffectingElement(&lmi_affected,o);

        _type = g_variant_get_byte(lmi_job_get_in_param(job, PARAM_TYPE));
        if (_type == JOB_RESTORE_LABELS || _type == JOB_SET_FILE_LABEL) {
            LMI_UnixFileRef lmi_ufr;
            const char *path = g_variant_get_string(lmi_job_get_in_param(job, "Path"), NULL);

            LMI_UnixFileRef_Init(&lmi_ufr, _cb, ns);
            LMI_UnixFileRef_Set_LFName(&lmi_ufr, path);
            LMI_UnixFileRef_Set_CSCreationClassName(&lmi_ufr, lmi_get_system_creation_class_name());
            LMI_UnixFileRef_Set_CSName(&lmi_ufr, lmi_get_system_name_safe(cc));
            o = LMI_UnixFileRef_ToObjectPath(&lmi_ufr, NULL);
            LMI_AffectedSELinuxJobElement_SetObjectPath_AffectedElement(&lmi_affected,o);
        } else {
            LMI_SELinuxElement lmi_elem;
            LMI_SELinuxElement_Init(&lmi_elem, _cb, ns);
            instance_id = g_variant_get_string(lmi_job_get_in_param(job, PARAM_INSTANCE_ID), NULL);
            LMI_SELinuxElement_Set_InstanceID(&lmi_elem, instance_id);
            o = LMI_SELinuxElement_ToObjectPath(&lmi_elem, NULL);
            LMI_AffectedSELinuxJobElement_SetObjectPath_AffectedElement(&lmi_affected,o);
        }

        KReturnInstance(cr, lmi_affected);
    }

    g_free(jobnums);
    KSetStatus(&st, OK);
    return st;
}

static CMPIStatus LMI_AffectedSELinuxJobElementGetInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_AffectedSELinuxJobElementCreateInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_AffectedSELinuxJobElementModifyInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci,
    const char **properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_AffectedSELinuxJobElementDeleteInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_AffectedSELinuxJobElementExecQuery(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *lang,
    const char *query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_AffectedSELinuxJobElementAssociationCleanup(
    CMPIAssociationMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_AffectedSELinuxJobElementAssociators(
    CMPIAssociationMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *assocClass,
    const char *resultClass,
    const char *role,
    const char *resultRole,
    const char **properties)
{
    return KDefaultAssociators(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_AffectedSELinuxJobElement_ClassName,
        assocClass,
        resultClass,
        role,
        resultRole,
        properties);
}

static CMPIStatus LMI_AffectedSELinuxJobElementAssociatorNames(
    CMPIAssociationMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *assocClass,
    const char *resultClass,
    const char *role,
    const char *resultRole)
{
    return KDefaultAssociatorNames(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_AffectedSELinuxJobElement_ClassName,
        assocClass,
        resultClass,
        role,
        resultRole);
}

static CMPIStatus LMI_AffectedSELinuxJobElementReferences(
    CMPIAssociationMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *assocClass,
    const char *role,
    const char **properties)
{
    return KDefaultReferences(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_AffectedSELinuxJobElement_ClassName,
        assocClass,
        role,
        properties);
}

static CMPIStatus LMI_AffectedSELinuxJobElementReferenceNames(
    CMPIAssociationMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *assocClass,
    const char *role)
{
    return KDefaultReferenceNames(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_AffectedSELinuxJobElement_ClassName,
        assocClass,
        role);
}

CMInstanceMIStub(
    LMI_AffectedSELinuxJobElement,
    LMI_AffectedSELinuxJobElement,
    _cb,
    LMI_AffectedSELinuxJobElementInitialize(ctx))

CMAssociationMIStub(
    LMI_AffectedSELinuxJobElement,
    LMI_AffectedSELinuxJobElement,
    _cb,
    LMI_AffectedSELinuxJobElementInitialize(ctx))

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_AffectedSELinuxJobElement",
    "LMI_AffectedSELinuxJobElement",
    "instance association")
