/*
 * Copyright (C) 2014 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Jan Synacek <jsynacek@redhat.com>
 *
 */

#include <konkret/konkret.h>
#include "LMI_AssociatedSELinuxJobMethodResult.h"
#include "selinux.h"

static const CMPIBroker *_cb;

static void LMI_AssociatedSELinuxJobMethodResultInitialize(const CMPIContext *ctx)
{
    selinux_provider_init(LMI_AssociatedSELinuxJobMethodResult_ClassName, FALSE, provider_config_defaults, _cb, ctx);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultCleanup(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    return jobmgr_cleanup(LMI_AssociatedSELinuxJobMethodResult_ClassName);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultEnumInstanceNames(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultEnumInstances(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    CMPIStatus status = {CMPI_RC_OK, NULL};
    CMPIObjectPath *jop, *mrop;
    LmiJob *job = NULL;
    guint *jobnum;
    gchar *jobid = NULL;
    guint *jobnums = jobmgr_get_job_numbers(NULL);

    if (jobnums == NULL)
        goto err;

    for (jobnum = jobnums; *jobnum != 0; ++jobnum) {
        job = jobmgr_get_job_by_number(*jobnum);
        if (!job) {
            /* job may have been deleted since numbers query */
            continue;
        }

        if (lmi_job_get_method_name(job) == NULL) {
            /* Method result can not be created for job with MethodName unset.
             * Such a job is too new. */
            goto next;
        }

        jobid = lmi_job_get_jobid(job);

        status = jobmgr_job_to_cim_op(job, &jop);
        if (status.rc) {
            if (status.msg) {
                lmi_warn("Failed to make job object path out of job \"%s\": %s",
                         jobid, CMGetCharsPtr(status.msg, NULL));
            } else {
                lmi_warn("Failed to make job object path out of job \"%s\"", jobid);
            }
            goto next;
        }

        status = jobmgr_job_to_method_result_op(job, NULL, &mrop);
        if (status.rc) {
            if (status.msg) {
                lmi_warn("Failed to make method result object path"
                         " out of job \"%s\": %s", jobid,
                         CMGetCharsPtr(status.msg, NULL));
            } else {
                lmi_warn("Failed to make method result object path"
                         " out of job \"%s\"", jobid);
            }
            CMRelease(jop);
            goto next;
        }

        LMI_AssociatedSELinuxJobMethodResult w;
        LMI_AssociatedSELinuxJobMethodResult_Init(&w, _cb, KNameSpace(cop));

        LMI_AssociatedSELinuxJobMethodResult_SetObjectPath_Job(&w, jop);
        LMI_AssociatedSELinuxJobMethodResult_SetObjectPath_JobParameters(&w, mrop);

        KReturnInstance(cr, w);
next:
        g_free(jobid);
        jobid = NULL;
        g_clear_object(&job);
    }
    g_free(jobnums);

    return status;

err:
    if (!status.rc) {
        lmi_error("Memory allocation failed");
        KSetStatus(&status, ERR_FAILED);
    }
    return status;
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultGetInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultCreateInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultModifyInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci,
    const char**properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultDeleteInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultExecQuery(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *lang,
    const char *query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultAssociationCleanup(
    CMPIAssociationMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultAssociators(
    CMPIAssociationMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *assocClass,
    const char *resultClass,
    const char *role,
    const char *resultRole,
    const char **properties)
{
    return KDefaultAssociators(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_AssociatedSELinuxJobMethodResult_ClassName,
        assocClass,
        resultClass,
        role,
        resultRole,
        properties);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultAssociatorNames(
    CMPIAssociationMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *assocClass,
    const char *resultClass,
    const char *role,
    const char *resultRole)
{
    return KDefaultAssociatorNames(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_AssociatedSELinuxJobMethodResult_ClassName,
        assocClass,
        resultClass,
        role,
        resultRole);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultReferences(
    CMPIAssociationMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *assocClass,
    const char *role,
    const char **properties)
{
    return KDefaultReferences(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_AssociatedSELinuxJobMethodResult_ClassName,
        assocClass,
        role,
        properties);
}

static CMPIStatus LMI_AssociatedSELinuxJobMethodResultReferenceNames(
    CMPIAssociationMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *assocClass,
    const char *role)
{
    return KDefaultReferenceNames(
        _cb,
        mi,
        cc,
        cr,
        cop,
        LMI_AssociatedSELinuxJobMethodResult_ClassName,
        assocClass,
        role);
}

CMInstanceMIStub(
    LMI_AssociatedSELinuxJobMethodResult,
    LMI_AssociatedSELinuxJobMethodResult,
    _cb,
    LMI_AssociatedSELinuxJobMethodResultInitialize(ctx))

CMAssociationMIStub(
    LMI_AssociatedSELinuxJobMethodResult,
    LMI_AssociatedSELinuxJobMethodResult,
    _cb,
    LMI_AssociatedSELinuxJobMethodResultInitialize(ctx))

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_AssociatedSELinuxJobMethodResult",
    "LMI_AssociatedSELinuxJobMethodResult",
    "instance association")
