/*
 * Copyright (C) 2014 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Jan Synacek <jsynacek@redhat.com>
 *
 * Implementation note:
 *
 * Casts to/from char/gchar to/from xmlChar are necessary, because xmlChar is
 * defined as an unsigned char.
 */

#include <konkret/konkret.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include "LMI_SELinuxBoolean.h"
#include "selinux.h"
#include <glib.h>

/**
 * Populate boolean description hash with descriptions of all SELinux booleans
 * specified by an XPath expression.
 *
 * This function parses /usr/share/selinux/devel/policy.xml to get boolean descriptions.
 * There's currently no way how to get this information via libselinux.
 *
 * @param desc_hash GHashTable pointer to the boolean description hash.
 * @param xpath_str XPath expression.
 * @retval 0 on success, otherwise -1.
 */
static int boolean_desc_hash_add_from_xpath(GHashTable *desc_hash, const xmlChar *xpath_str)
{
    xmlDoc *doc = NULL;
    xmlXPathContextPtr xpath_ctx = NULL;
    xmlXPathObjectPtr xpath_obj = NULL;
    xmlNodeSetPtr nodes;
    xmlNodePtr p_node;

    doc = xmlParseFile("/usr/share/selinux/devel/policy.xml");
    if (doc == NULL) {
        lmi_warn("Couldn't read SELinux policy");
        return -1;
    }

    xpath_ctx = xmlXPathNewContext(doc);
    if(xpath_ctx == NULL) {
        lmi_warn("Unable to create new XPath context");
        xmlFreeDoc(doc);
        return -1;
    }

    xpath_obj = xmlXPathEvalExpression(xpath_str, xpath_ctx);
    if(xpath_obj == NULL) {
        lmi_warn("Unable to evaluate xpath expression: %s", xpath_str);
        xmlXPathFreeContext(xpath_ctx);
        xmlFreeDoc(doc);
        return -1;
    }

    xmlChar *parent_name;
    xmlChar *content;
    nodes = xpath_obj->nodesetval;
    for(int i = 0; i < nodes->nodeNr; ++i) {
        if(nodes->nodeTab[i]->type == XML_ELEMENT_NODE) {
            p_node = nodes->nodeTab[i];
            parent_name = xmlGetProp(p_node->parent->parent, (const xmlChar *)"name");
            content = p_node->children->content;

            lmi_info("Adding SELinux boolean to hash: %s = %s", parent_name, content);
            g_hash_table_insert(desc_hash,
                                g_strdup((gchar *)parent_name),
                                lmi_strip_str((gchar *)content, '\n'));

            xmlFree(parent_name);
        }
    }

    xmlXPathFreeObject(xpath_obj);
    xmlXPathFreeContext(xpath_ctx);
    xmlFreeDoc(doc);
    return 0;
}

/**
 * Look up a SELinux boolean description.
 *
 * @param desc_hash GHashTable pointer to the boolean description hash.
 * @param boolean Boolean name, description of which will be returned.
 * @retval Boolean description.
 */
static const char *boolean_desc_hash_lookup(GHashTable *desc_hash, const char *boolean)
{
    return g_hash_table_lookup(desc_hash, boolean);
}

/**
 * Initialize a hash table containing SELinux boolean descriptions.
 * If anything fails, NULL is returned.
 *
 * @retval GHashTable pointer to the initialized hash table, or NULL.
 */
static GHashTable *boolean_desc_hash_init()
{
    int rc;
    GHashTable *desc_hash = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
    if (desc_hash == NULL) return NULL;

    rc = boolean_desc_hash_add_from_xpath(desc_hash, (const xmlChar *)"/policy/layer/module/tunable/desc/p");
    if (rc < 0) return NULL;
    rc = boolean_desc_hash_add_from_xpath(desc_hash, (const xmlChar *)"/policy/layer/module/bool/desc/p");
    if (rc < 0) return NULL;
    rc = boolean_desc_hash_add_from_xpath(desc_hash, (const xmlChar *)"/policy/layer/bool/desc/p");
    if (rc < 0) return NULL;
    rc = boolean_desc_hash_add_from_xpath(desc_hash, (const xmlChar *)"/policy/tunable/desc/p");
    if (rc < 0) return NULL;

    return desc_hash;
}

/**
 * Free SELinux boolean description hash table.
 *
 * @param desc_hash GHashTable pointer to the boolean description hash.
 */
static void boolean_desc_hash_free(GHashTable *desc_hash)
{
    g_hash_table_unref(desc_hash);
}

static const CMPIBroker *_cb = NULL;

static void LMI_SELinuxBooleanInitialize(const CMPIContext *ctx)
{
    lmi_info("Initializing XML parser.");
    xmlInitParser();
    lmi_init(provider_name, _cb, ctx, provider_config_defaults);
}

static CMPIStatus LMI_SELinuxBooleanCleanup(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    lmi_info("Cleaning up XML parser.");
    xmlCleanupParser();
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_SELinuxBooleanEnumInstanceNames(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}


static CMPIStatus LMI_SELinuxBooleanEnumInstances(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    const char *ns = KNameSpace(cop);
    int rc;
    int len;
    char **booleans;
    GHashTable *b_hash = boolean_desc_hash_init();
    const char *b_val;
    semanage_handle_t *hnd = NULL;
    semanage_bool_key_t *bk = NULL;
    semanage_bool_t *b = NULL;

    rc = security_get_boolean_names(&booleans, &len);
    /* this function is useless without available booleans, so make error fatal */
    if (rc < 0) {
        if (b_hash)
            boolean_desc_hash_free(b_hash);
        lmi_return_with_chars(_cb, CMPI_RC_ERR_FAILED, "Could not get SELinux booleans");
    }

    hnd = semanage_handle_create();
    if (!hnd) {
        lmi_error("semanage_handle_create failed");
        lmi_return_with_chars(_cb, CMPI_RC_ERR_FAILED, "Could not get SELinux booleans");
    }

    rc = semanage_connect(hnd);
    if (rc < 0) {
        lmi_error("semanage_connect failed");
        lmi_return_with_chars(_cb, CMPI_RC_ERR_FAILED, "Could not get SELinux booleans");
    }

    for (int i = 0; i < len; i++) {
        LMI_SELinuxBoolean lmi_bool;

        char instance_id[BUFLEN];
        snprintf(instance_id, BUFLEN,
                 LMI_ORGID ":" LMI_SELinuxBoolean_ClassName ":%s", booleans[i]);

        rc = semanage_bool_key_create(hnd, booleans[i], &bk);
        if (rc < 0) {
            lmi_error("semanage_bool_key_create failed: %s", booleans[i]);
            continue;
        }
        rc = semanage_bool_query_local(hnd, bk, &b);
        if (rc < 0) {
            rc = semanage_bool_query(hnd, bk, &b);
            if (rc < 0) {
                lmi_error("semanage_bool_query failed: %s", booleans[i]);
                continue;
            }
        }

        LMI_SELinuxBoolean_Init(&lmi_bool, _cb, ns);
        LMI_SELinuxBoolean_Set_InstanceID(&lmi_bool, instance_id);
        LMI_SELinuxBoolean_Set_ElementName(&lmi_bool, booleans[i]);
        LMI_SELinuxBoolean_Set_State(&lmi_bool, security_get_boolean_active(booleans[i]));
        LMI_SELinuxBoolean_Set_DefaultState(&lmi_bool, semanage_bool_get_value(b));
        b_val = (b_hash) ? boolean_desc_hash_lookup(b_hash, booleans[i]) : "N/A";
        LMI_SELinuxBoolean_Set_Description(&lmi_bool, b_val);

        KReturnInstance(cr, lmi_bool);
        semanage_bool_key_free(bk);
        semanage_bool_free(b);
    }

    for (int i=0; i< len; i++)
        free(booleans[i]);
    free(booleans);

    if (b_hash)
        boolean_desc_hash_free(b_hash);
    semanage_disconnect(hnd);
    semanage_handle_destroy(hnd);
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_SELinuxBooleanGetInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_SELinuxBooleanCreateInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxBooleanModifyInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci,
    const char **properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxBooleanDeleteInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxBooleanExecQuery(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *lang,
    const char *query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

CMInstanceMIStub(
    LMI_SELinuxBoolean,
    LMI_SELinuxBoolean,
    _cb,
    LMI_SELinuxBooleanInitialize(ctx))

static CMPIStatus LMI_SELinuxBooleanMethodCleanup(
    CMPIMethodMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_SELinuxBooleanInvokeMethod(
    CMPIMethodMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *meth,
    const CMPIArgs *in,
    CMPIArgs *out)
{
    return LMI_SELinuxBoolean_DispatchMethod(
        _cb, mi, cc, cr, cop, meth, in, out);
}

CMMethodMIStub(
    LMI_SELinuxBoolean,
    LMI_SELinuxBoolean,
    _cb,
    LMI_SELinuxBooleanInitialize(ctx))

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_SELinuxBoolean",
    "LMI_SELinuxBoolean",
    "instance method")
