/*
 * Copyright (C) 2014 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Jan Synacek <jsynacek@redhat.com>
 *
 */

#include <konkret/konkret.h>
#include "LMI_SELinuxJob.h"
#include "selinux.h"
#include "job_manager.h"

static const CMPIBroker *_cb;

static void LMI_SELinuxJobInitialize(const CMPIContext *ctx)
{
    selinux_provider_init(LMI_SELinuxJob_ClassName, FALSE, provider_config_defaults, _cb, ctx);
}

static CMPIStatus LMI_SELinuxJobCleanup(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    return jobmgr_cleanup(LMI_SELinuxJob_ClassName);
}

static CMPIStatus LMI_SELinuxJobEnumInstanceNames(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_SELinuxJobEnumInstances(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    CMPIStatus status = {CMPI_RC_OK, NULL};
    gchar *jobid;
    LmiJob *job;
    CMPIInstance *inst;
    guint *jobnum;
    guint *jobnums = jobmgr_get_job_numbers(NULL);

    if (jobnums == NULL)
        CMReturn(CMPI_RC_ERR_FAILED);

    for (jobnum = jobnums; *jobnum != 0; jobnum++) {
        job = jobmgr_get_job_by_number(*jobnum);
        if (!job)
            /* job may have been deleted since numbers query */
            continue;

        if (!LMI_IS_JOB(job)) {
            g_object_unref(job);
            continue;
        }


        status = jobmgr_job_to_cim_instance(job, &inst);
        g_object_unref(job);
        if (status.rc != CMPI_RC_OK) {
            jobid = lmi_job_get_jobid(job);
            if (status.msg) {
                lmi_error("Failed to make cim instance out of job \"%s\": %s",
                               jobid, CMGetCharsPtr(status.msg, NULL));
            } else {
                lmi_error("Failed to make cim instance out of job \"%s\"", jobid);
            }
            g_free(jobid);
            continue;
        }

        CMReturnInstance(cr, inst);
    }

    g_free(jobnums);

    KSetStatus(&status, OK);
    return status;
}

static CMPIStatus LMI_SELinuxJobGetInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_SELinuxJobCreateInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxJobModifyInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci,
    const char **properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxJobDeleteInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxJobExecQuery(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *lang,
    const char *query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

CMInstanceMIStub(
    LMI_SELinuxJob,
    LMI_SELinuxJob,
    _cb,
    LMI_SELinuxJobInitialize(ctx))

static CMPIStatus LMI_SELinuxJobMethodCleanup(
    CMPIMethodMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_SELinuxJobInvokeMethod(
    CMPIMethodMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *meth,
    const CMPIArgs *in,
    CMPIArgs *out)
{
    return LMI_SELinuxJob_DispatchMethod(
        _cb, mi, cc, cr, cop, meth, in, out);
}

CMMethodMIStub(
    LMI_SELinuxJob,
    LMI_SELinuxJob,
    _cb,
    LMI_SELinuxJobInitialize(ctx))

KUint32 LMI_SELinuxJob_KillJob(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxJobRef *self,
    const KBoolean *DeleteOnKill,
    CMPIStatus *status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_SELinuxJob_RequestStateChange(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxJobRef *self,
    const KUint16 *RequestedState,
    const KDateTime *TimeoutPeriod,
    CMPIStatus *status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_SELinuxJob_GetError(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxJobRef *self,
    KInstance *Error,
    CMPIStatus *status)
{
    CMPIObjectPath *cop;
    LmiJob *job = NULL;
    CMPIInstance *inst = NULL;
    KUint32 result = KUINT32_INIT;

    if ((cop = LMI_SELinuxJobRef_ToObjectPath(self, status)) == NULL)
        goto err;
    *status = jobmgr_get_job_matching_op(cop, &job);
    CMRelease(cop);
    if (status->rc) {
        KSetStatus(status, ERR_NOT_FOUND);
        goto done;
    }
    JOB_CRITICAL_BEGIN(job);
    if (lmi_job_get_state(job) == LMI_JOB_STATE_ENUM_EXCEPTION) {
        *status = jobmgr_job_to_cim_error(job, &inst);
        if (status->rc)
            goto critical_end_err;
        KInstance_Set(Error, inst);
    }
    JOB_CRITICAL_END(job);
    g_clear_object(&job);

    KUint32_Set(&result, 0);
    goto done;

critical_end_err:
    JOB_CRITICAL_END(job);
    g_clear_object(&job);
err:
    if (!status->rc) {
        lmi_error("Memory allocation failed");
        KSetStatus(status, ERR_FAILED);
    }
done:
    return result;
}

KUint32 LMI_SELinuxJob_GetErrors(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxJobRef *self,
    KInstanceA *errors,
    CMPIStatus *status)
{
    KUint32 result = KUINT32_INIT;
    KInstance error = KINSTANCE_INIT;

    KInstance_Null(&error);
    LMI_SELinuxJob_GetError(cb, mi, context, self, &error, status);
    if (status->rc)
        return result;
    if (KHasValue(&error)) {
        KInstanceA_Init(errors, cb, 1);
        KInstanceA_Set(errors, 0, (CMPIInstance *) error.value);
    } else {
        KInstanceA_Init(errors, cb, 0);
    }
    KUint32_Set(&result, 0);
    return result;
}

KUint32 LMI_SELinuxJob_ResumeWithAction(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxJobRef *self,
    CMPIStatus *status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_SELinuxJob_ResumeWithInput(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxJobRef *self,
    const KStringA *Inputs,
    CMPIStatus *status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(status, ERR_NOT_SUPPORTED);
    return result;
}

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_SELinuxJob",
    "LMI_SELinuxJob",
    "instance method")
