/*
 * Copyright (C) 2014 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Jan Synacek <jsynacek@redhat.com>
 *
 */

#include <konkret/konkret.h>
#include "LMI_SELinuxPort.h"
#include "selinux.h"

static const CMPIBroker *_cb = NULL;

/**
 * Construct a new string from a port instance.
 *
 * Libsemanage returns port ranges as ints.
 *
 * @param port Libsemanage port.
 * @retval Newly allocated string with a port range. Must be freed by the caller.
 */
static char *get_port_range_str(const semanage_port_t *port)
{
    int portnum_l, portnum_h;
    char *result;

    portnum_l = semanage_port_get_low(port);
    portnum_h = semanage_port_get_high(port);
    if (portnum_l == portnum_h) {
        if (asprintf(&result, "%d", portnum_l) < 0) {
            return NULL;
        }
    }
    else {
        if (asprintf(&result, "%d-%d", portnum_l, portnum_h) < 0) {
            return NULL;
        }
    }

    return result;
}

/**
 * Handler to be executed for each port within the port hash.
 *
 * Creates LMI_SELinuxPort instances and sends them to the broker.
 *
 * @param key InstanceID of soon-to-be LMI_SELinuxPort.
 * @param value Linked list containing all libsemanage ports associated with the InstanceID.
 * @param user_data CMPI context data.
 */
static void port_hash_foreach_handler(gpointer key, gpointer value, gpointer user_data)
{
    const struct cmpi_data *data = (const struct cmpi_data *)user_data;
    const CMPIResult *cr = data->cr;
    const char *ns = data->namespace;

    semanage_context_t *ctx;
    const char *type, *role, *user;
    char contextstr[BUFLEN];

    const char *instance_id = (const char *)key;
    GSList *ports = (GSList *)value;
    semanage_port_t *port = (semanage_port_t *)g_slist_nth_data(ports, 0);
    int n_ports = g_slist_length(ports);

    ctx = semanage_port_get_con(port);
    type = semanage_context_get_type(ctx);
    role = semanage_context_get_role(ctx);
    user = semanage_context_get_user(ctx);
    snprintf(contextstr, BUFLEN, "%s:%s:%s", user, role, type);

    LMI_SELinuxPort lmi_port;
    LMI_SELinuxPort_Init(&lmi_port, _cb, ns);
    LMI_SELinuxPort_Set_InstanceID(&lmi_port, instance_id);
    LMI_SELinuxPort_Set_ElementName(&lmi_port, type);
    LMI_SELinuxPort_Set_Protocol(&lmi_port, semanage_port_get_proto(port));
    LMI_SELinuxPort_Set_SELinuxContext(&lmi_port, contextstr);

    LMI_SELinuxPort_Init_Ports(&lmi_port, n_ports);
    for (int i = 0; i < n_ports; i++) {
        semanage_port_t *port = (semanage_port_t *)g_slist_nth_data(ports, i);
        char *p_str = get_port_range_str(port);
        LMI_SELinuxPort_Set_Ports(&lmi_port, i, p_str);
        free(p_str);
    }

    CMReturnInstance(cr, LMI_SELinuxPort_ToInstance(&lmi_port, NULL));
}

static void port_free(gpointer data)
{
    semanage_port_free((semanage_port_t *)data);
}

static gboolean port_hash_free(gpointer key, gpointer value, gpointer user_data)
{
    g_slist_free_full((GSList *)value, port_free);
    return TRUE;
}

static void LMI_SELinuxPortInitialize(const CMPIContext *ctx)
{
    lmi_init(provider_name, _cb, ctx, provider_config_defaults);
}

static CMPIStatus LMI_SELinuxPortCleanup(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_SELinuxPortEnumInstanceNames(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_SELinuxPortEnumInstances(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    const char *ns = KNameSpace(cop);
    int rc;

    semanage_handle_t *manage_hnd = NULL;
    manage_hnd = semanage_handle_create();
    if (manage_hnd == NULL) {
        lmi_return_with_chars(_cb, CMPI_RC_ERR_FAILED, "Could not create semanage handle");
    }

    rc = semanage_connect(manage_hnd);
    if (rc < 0) {
        lmi_return_with_chars(_cb, CMPI_RC_ERR_FAILED, "Could not connect to semanage");
    }

    struct cmpi_data data = {cr, ns};
    semanage_port_t **ports = NULL;
    unsigned int n_ports;
    unsigned int i = 0;
    rc = semanage_port_list(manage_hnd, &ports, &n_ports);
    if (rc < 0) {
        semanage_disconnect(manage_hnd);
        semanage_handle_destroy(manage_hnd);
        lmi_return_with_chars(_cb, CMPI_RC_ERR_FAILED, "Could not list SELinux ports");
    }
    /*
      Libsemanage returns ports in the following manner:
      ...
      system_u:object_r:clockspeed_port_t       4041/udp
      system_u:object_r:cluster_port_t          5149/tcp
      system_u:object_r:cluster_port_t          5149/udp
      system_u:object_r:cluster_port_t         40040/tcp
      system_u:object_r:cluster_port_t   50006-50008/tcp
      system_u:object_r:cluster_port_t   50006-50008/udp
      system_u:object_r:cma_port_t              1050/tcp
      system_u:object_r:cma_port_t              1050/udp
      ...

      It is currently not possible to get port ranges for particular SELinux
      ports in a sensible way, therefore the complex code below exists.

      It works in 3 passes. First, ports are iterated one by one (as is the only
      possible way to go over them) and a hash is built, using port instance id
      as a key, and a list of port instances as its value. The port instances
      differ only in the port range values, which is why it is necessary to
      process them like this. Second pass processes the hash and builds and
      returns class instances to the broker. Third pass frees the allocated
      ports. This step is, in theory, not necessary, but it's cleaner to free
      ports in a dedicated step, and not as a part of the previous algorithm.

      On my testing system, this makes 556 (ports returned by libsemanage) + 381
      (unique ports from the hash) + 556 (ports to be freed) iterations, which
      makes it roughly 3 times slower than it could be, if libsemanage returned
      its data in a sensible way.
    */

    /* First pass: reorganize in a sensible way. */
    GHashTable *ports_hash;
    /* The values (lists containing the libsemanage port structs) don't have to
     * be freed, as the ports will be freed separately in the third pass. */
    ports_hash = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);

    for (; i < n_ports; i++) {
        semanage_context_t *ctx = semanage_port_get_con(ports[i]);
        const char *p_type = semanage_context_get_type(ctx);
        const char *p_proto = PORT_PROTO(ports[i]);
        GSList *ranges;

        char instance_id[BUFLEN];
        snprintf(instance_id, BUFLEN,
                 LMI_ORGID ":" LMI_SELinuxPort_ClassName ":%s:%s", p_proto, p_type);

        /* Lookup the value first. If it is found, the list can be extended and
         * the original pointer replaced. If not found, the NULL will be used,
         * indicating an empty list. */
        ranges = g_hash_table_lookup(ports_hash, instance_id);
        ranges = g_slist_prepend(ranges, ports[i]);
        g_hash_table_insert(ports_hash, g_strdup(instance_id), ranges);
    }

    /* Second pass: feed the broker. */
    g_hash_table_foreach(ports_hash, port_hash_foreach_handler, &data);

    /* Third pass: free ports and lists that hold them. */
    g_hash_table_foreach_remove(ports_hash, port_hash_free, NULL);

    if (ports)
        free(ports);
    g_hash_table_destroy(ports_hash);
    semanage_disconnect(manage_hnd);
    semanage_handle_destroy(manage_hnd);
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_SELinuxPortGetInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_SELinuxPortCreateInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxPortModifyInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci,
    const char **properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxPortDeleteInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxPortExecQuery(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *lang,
    const char *query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

CMInstanceMIStub(
    LMI_SELinuxPort,
    LMI_SELinuxPort,
    _cb,
    LMI_SELinuxPortInitialize(ctx))

static CMPIStatus LMI_SELinuxPortMethodCleanup(
    CMPIMethodMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_SELinuxPortInvokeMethod(
    CMPIMethodMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *meth,
    const CMPIArgs *in,
    CMPIArgs *out)
{
    return LMI_SELinuxPort_DispatchMethod(
        _cb, mi, cc, cr, cop, meth, in, out);
}

CMMethodMIStub(
    LMI_SELinuxPort,
    LMI_SELinuxPort,
    _cb,
    LMI_SELinuxPortInitialize(ctx))

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_SELinuxPort",
    "LMI_SELinuxPort",
    "instance method")
