/*
 * Copyright (C) 2014 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Jan Synacek <jsynacek@redhat.com>
 *
 */

#include <konkret/konkret.h>
#include "LMI_SELinuxService.h"
#include "selinux.h"
#include <assert.h>

static const CMPIBroker *_cb = NULL;



/**
 * Parse SELinux policy type from the SELinux configuration.
 *
 * @param policy_type SELinux policy type. Must be freed by the caller.
 */
static void parse_selinux_sysconf(char **policy_type)
{
    const char *conf_path = "/etc/selinux/config";
    const char *fake_group = "fake";
    const char *fake_group_str = "[fake]\n";
    const int fake_group_str_len = strlen(fake_group_str);
    off_t size = 0;
    gchar *content;
    GError *error;
    GKeyFile *keyfile;
    FILE *f;

    lmi_debug("==> parse_selinux_sysconf (file=%s)", conf_path);
    f = open_selinux_config("r", &size);
    size += fake_group_str_len;

    content = g_malloc0(size + 1);
    if (content == NULL) {
        lmi_error("Memory allocation failed");
        return;
    }

    strcat(content, fake_group_str);
    if (f != NULL) {
        fread(content + fake_group_str_len, 1, size - fake_group_str_len, f);
        fclose(f);
    }

    keyfile = g_key_file_new();

    if (!g_key_file_load_from_data(keyfile, content, size,
                                   G_KEY_FILE_NONE, &error))
    {
        lmi_error("Can't read SELinux configuration: %s", error->message);
        g_key_file_free(keyfile);
        g_free(content);
        return;
    }

    *policy_type = g_key_file_get_value(keyfile, fake_group, "SELINUXTYPE", &error);

    g_key_file_free(keyfile);
    g_free(content);
    lmi_debug("<== parse_selinux_sysconf");
}



static void LMI_SELinuxServiceInitialize(const CMPIContext *ctx)
{
    selinux_provider_init(LMI_SELinuxService_ClassName, FALSE, provider_config_defaults, _cb, ctx);
}


static CMPIStatus LMI_SELinuxServiceCleanup(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    return jobmgr_cleanup(LMI_SELinuxService_ClassName);
}

static CMPIStatus LMI_SELinuxServiceEnumInstanceNames(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    return KDefaultEnumerateInstanceNames(
        _cb, mi, cc, cr, cop);
}

static CMPIStatus LMI_SELinuxServiceEnumInstances(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    const char *ns = KNameSpace(cop);
    char *pt = NULL;
    int ds;
    LMI_SELinuxService lmi_service;

    init_selinux_service(&lmi_service, _cb, cc, ns);
    LMI_SELinuxService_Set_PolicyVersion(&lmi_service, security_policyvers());
    /* security_getenforce() / security_getenforcemode() return -1 (disabled), 0
     * (permissive) and 1 (enforcing).  If +1 is added to the value, the correct
     * value from the CMPI value map is obtained (see MOF for details). */
    LMI_SELinuxService_Set_SELinuxState(&lmi_service, security_getenforce() + 1);

    parse_selinux_sysconf(&pt);
    selinux_getenforcemode(&ds);
    LMI_SELinuxService_Set_PolicyType(&lmi_service, pt);
    LMI_SELinuxService_Set_SELinuxDefaultState(&lmi_service, ds + 1);

    if (pt)
        free(pt);
    KReturnInstance(cr, lmi_service);
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_SELinuxServiceGetInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char **properties)
{
    return KDefaultGetInstance(
        _cb, mi, cc, cr, cop, properties);
}

static CMPIStatus LMI_SELinuxServiceCreateInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxServiceModifyInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const CMPIInstance *ci,
    const char **properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxServiceDeleteInstance(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SELinuxServiceExecQuery(
    CMPIInstanceMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *lang,
    const char *query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

CMInstanceMIStub(
    LMI_SELinuxService,
    LMI_SELinuxService,
    _cb,
    LMI_SELinuxServiceInitialize(ctx))

static CMPIStatus LMI_SELinuxServiceMethodCleanup(
    CMPIMethodMI *mi,
    const CMPIContext *cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus LMI_SELinuxServiceInvokeMethod(
    CMPIMethodMI *mi,
    const CMPIContext *cc,
    const CMPIResult *cr,
    const CMPIObjectPath *cop,
    const char *meth,
    const CMPIArgs *in,
    CMPIArgs *out)
{
    return LMI_SELinuxService_DispatchMethod(
        _cb, mi, cc, cr, cop, meth, in, out);
}

CMMethodMIStub(
    LMI_SELinuxService,
    LMI_SELinuxService,
    _cb,
    LMI_SELinuxServiceInitialize(ctx))

KUint32 LMI_SELinuxService_RequestStateChange(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxServiceRef *self,
    const KUint16 *RequestedState,
    KRef *Job,
    const KDateTime *TimeoutPeriod,
    CMPIStatus *__status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(__status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_SELinuxService_StartService(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxServiceRef *self,
    CMPIStatus *__status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(__status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_SELinuxService_StopService(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxServiceRef *self,
    CMPIStatus *__status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(__status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_SELinuxService_ChangeAffectedElementsAssignedSequence(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxServiceRef *self,
    const KRefA *ManagedElements,
    const KUint16A *AssignedSequence,
    KRef *Job,
    CMPIStatus *__status)
{
    KUint32 result = KUINT32_INIT;

    KSetStatus(__status, ERR_NOT_SUPPORTED);
    return result;
}

KUint32 LMI_SELinuxService_SetSELinuxState(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxServiceRef *self,
    const KUint16 *NewState,
    const KBoolean *MakeDefault,
    KRef *Job,
    CMPIStatus *__status)
{
    KUint32 result = KUINT32_INIT;
    CMPIObjectPath *cop;
    LmiJob *job;
    GVariant *v;

    if (!KHasValue(NewState)) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, LMI_PARAM_ERROR_STR(NewState));
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }
    if (NewState->value > 2) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, "NewState out of range");
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    /* MakeDefault defaults to False */

    /* setenforce doesn't allow to disable SELinux in runtime, the provider shouldn't either */
    if (KHasValue(MakeDefault) && MakeDefault->value == 0 && NewState->value == 0) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, "Disabling SELinux in runtime is not allowed");
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    job = jobmgr_new_job(LMI_TYPE_JOB);
    if (job == NULL) {
        KSetStatus2(cb, __status, ERR_FAILED, "Could not create job instance");
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }
    lmi_job_set_method_name(job, "SetSELinuxState");

    v = g_variant_new_int16(NewState->value);
    lmi_job_set_in_param(job, "NewState", v);
    v = g_variant_new_boolean(MakeDefault->value);
    lmi_job_set_in_param(job, "MakeDefault", v);
    v = g_variant_new_byte(JOB_SET_SELINUX_STATE);
    lmi_job_set_in_param(job, PARAM_TYPE, v);

    *__status = jobmgr_run_job(job);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    *__status = jobmgr_job_to_cim_op(job, &cop);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    KRef_SetObjectPath(Job, cop);
    KUint32_Set(&result, JOB_RC_STARTED);
    return result;
}

static CMPIStatus parse_port_instance_id(const gchar *instance_id, gint *type, gchar **name)
{
    CMPIStatus st = {0, NULL};
    gchar **parts;

    parts = g_strsplit(instance_id, ":", 0);
    if (!parts) {
        KSetStatus2(_cb, &st, ERR_INVALID_PARAMETER, "Port InstanceID could not be parsed");
        return st;
    }

    if (g_strv_length(parts) != 4) {
        KSetStatus2(_cb, &st, ERR_INVALID_PARAMETER, "Port InstanceID is not valid (wrong format)");
        g_strfreev(parts);
        return st;
    }
    if (!PORT_STR_IS_VALID(parts[2])) {
        KSetStatus2(_cb, &st, ERR_INVALID_PARAMETER, "Port InstanceID is not valid (invalid protocol)");
        g_strfreev(parts);
        return st;
    }

    *type = PORT_STR_TO_ENUM(parts[2]);
    *name = parts[3];

    g_free(parts[0]);
    g_free(parts[1]);
    g_free(parts[2]);
    g_free(parts);
    return st;
}

static CMPIStatus parse_port_range(const gchar *range, gint *low, gint *high)
{
    CMPIStatus st = {0, NULL};
    gchar **parts;
    gint parts_len;

    lmi_debug("==> parse_port_range (range=%s)", range);
    parts = g_strsplit(range, "-", 0);
    if (!parts) {
        KSetStatus2(_cb, &st, ERR_INVALID_PARAMETER, "PortRange could not be parsed");
        return st;
    }

    parts_len = g_strv_length(parts);
    if (parts_len > 2) {
        KSetStatus2(_cb, &st, ERR_INVALID_PARAMETER, "Invalid PortRange");
        goto out;
    }

    *low = atoi(parts[0]);
    *high = (parts_len == 2) ? atoi(parts[1]) : *low;
    if (*low == 0 || *high == 0) {
        KSetStatus2(_cb, &st, ERR_INVALID_PARAMETER, "Invalid PortRange");
        goto out;
    }
    lmi_debug("<== parse_port_range (low=%d, high=%d)", *low, *high);

out:
    g_strfreev(parts);
    return st;
}

KUint32 LMI_SELinuxService_SetPortLabel(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxServiceRef *self,
    const KRef *Target,
    const KString *PortRange,
    KRef *Job,
    CMPIStatus *__status)
{
    KUint32 result = KUINT32_INIT;
    CMPIObjectPath *cop;
    LmiJob *job;
    GVariant *v;
    LMI_SELinuxPortRef lmi_portr;
    gchar *port_name;
    gint port_type, range_low, range_high;
    const gchar *instance_id;

    if (!KHasValue(PortRange)) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, LMI_PARAM_ERROR_STR(PortRange));
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    if (!KHasValue(Target)) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, LMI_PARAM_ERROR_STR(Target));
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    *__status = LMI_SELinuxPortRef_InitFromObjectPath(&lmi_portr, cb, Target->value);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }
    if (!lmi_portr.InstanceID.chars || strcmp(lmi_portr.InstanceID.chars, "") == 0) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, "The Target misses InstanceID");
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    *__status = parse_port_instance_id(lmi_portr.InstanceID.chars, &port_type, &port_name);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    *__status = parse_port_range(PortRange->value->ft->getCharPtr(PortRange->value, NULL), &range_low, &range_high);
    if (__status->rc) {
        g_free(port_name);
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    job = jobmgr_new_job(LMI_TYPE_JOB);
    if (job == NULL) {
        g_free(port_name);
        KSetStatus2(cb, __status, ERR_FAILED, "Could not create job instance");
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }
    lmi_job_set_method_name(job, "SetPortLabel");

    v = g_variant_new_int32(port_type);
    lmi_job_set_in_param(job, "Type", v);
    v = g_variant_new_string(port_name);
    lmi_job_set_in_param(job, "Name", v);

    g_free(port_name);

    v = g_variant_new_int32(range_low);
    lmi_job_set_in_param(job, "RangeLow", v);
    v = g_variant_new_int32(range_high);
    lmi_job_set_in_param(job, "RangeHigh", v);
    v = g_variant_new_byte(JOB_SET_PORT_LABEL);
    lmi_job_set_in_param(job, PARAM_TYPE, v);
    instance_id = lmi_get_string_property_from_objectpath(Target->value, "InstanceID");
    v = g_variant_new_string(instance_id);
    lmi_job_set_in_param(job, PARAM_INSTANCE_ID, v);

    *__status = jobmgr_run_job(job);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    *__status = jobmgr_job_to_cim_op(job, &cop);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    KRef_SetObjectPath(Job, cop);
    KUint32_Set(&result, JOB_RC_STARTED);
    return result;
}

KUint32 LMI_SELinuxService_SetFileLabel(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxServiceRef *self,
    const KRef *Target,
    const KString *Label,
    KRef *Job,
    CMPIStatus *__status)
{
    KUint32 result = KUINT32_INIT;
    CMPIObjectPath *cop;
    LmiJob *job;
    GVariant *v;
    LMI_UnixFileRef lmi_ufr;
    CMPIString *str;
    const gchar *cstr;

    if (!KHasValue(Target)) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, LMI_PARAM_ERROR_STR(Target));
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    if (!KHasValue(Label)) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, LMI_PARAM_ERROR_STR(Label));
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    job = jobmgr_new_job(LMI_TYPE_JOB);
    if (job == NULL) {
        KSetStatus2(cb, __status, ERR_FAILED, "Could not create job instance");
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }
    lmi_job_set_method_name(job, "SetFileLabel");

    *__status = LMI_UnixFileRef_InitFromObjectPath(&lmi_ufr, cb, Target->value);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }
    if (!lmi_ufr.LFName.chars) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, "The Target misses LFName");
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    str = Label->value;
    cstr = str->ft->getCharPtr(str, NULL);
    if (!cstr || strcmp(cstr, "") == 0) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, "The Label cannot be empty");
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    v = g_variant_new_string(lmi_ufr.LFName.chars);
    lmi_job_set_in_param(job, "Path", v);
    v = g_variant_new_string(cstr);
    lmi_job_set_in_param(job, "Label", v);
    v = g_variant_new_byte(JOB_SET_FILE_LABEL);
    lmi_job_set_in_param(job, PARAM_TYPE, v);

    *__status = jobmgr_run_job(job);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    *__status = jobmgr_job_to_cim_op(job, &cop);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    KRef_SetObjectPath(Job, cop);
    KUint32_Set(&result, JOB_RC_STARTED);
    return result;
}

KUint32 LMI_SELinuxService_SetBoolean(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxServiceRef *self,
    const KRef *Target,
    const KBoolean *Value,
    const KBoolean *MakeDefault,
    KRef *Job,
    CMPIStatus *__status)
{
    KUint32 result = KUINT32_INIT;
    CMPIObjectPath *cop;
    LmiJob *job;
    GVariant *v;
    LMI_SELinuxBooleanRef lmi_boolr;
    gchar **parts;
    const gchar *instance_id;

    if (!KHasValue(Target)) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, LMI_PARAM_ERROR_STR(Target));
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    if (!KHasValue(Value)) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, LMI_PARAM_ERROR_STR(Value));
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    /* MakeDefault defaults to False */

    *__status = LMI_SELinuxBooleanRef_InitFromObjectPath(&lmi_boolr, cb, Target->value);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    parts = g_strsplit(lmi_boolr.InstanceID.chars, ":", 0);
    if (!parts) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, "Invalid InstanceID");
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    job = jobmgr_new_job(LMI_TYPE_JOB);
    if (job == NULL) {
        KSetStatus2(cb, __status, ERR_FAILED, "Could not create job instance");
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }
    lmi_job_set_method_name(job, "SetBoolean");

    v = g_variant_new_string(parts[2]);
    g_strfreev(parts);
    lmi_job_set_in_param(job, "BoolName", v);
    v = g_variant_new_boolean(Value->value);
    lmi_job_set_in_param(job, "Value", v);
    v = g_variant_new_boolean(MakeDefault->value);
    lmi_job_set_in_param(job, "MakeDefault", v);
    v = g_variant_new_byte(JOB_SET_BOOLEAN);
    lmi_job_set_in_param(job, PARAM_TYPE, v);
    instance_id = lmi_get_string_property_from_objectpath(Target->value, "InstanceID");
    v = g_variant_new_string(instance_id);
    lmi_job_set_in_param(job, PARAM_INSTANCE_ID, v);

    *__status = jobmgr_run_job(job);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    *__status = jobmgr_job_to_cim_op(job, &cop);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    KRef_SetObjectPath(Job, cop);
    KUint32_Set(&result, JOB_RC_STARTED);
    return result;
}

KUint32 LMI_SELinuxService_RestoreLabels(
    const CMPIBroker *cb,
    CMPIMethodMI *mi,
    const CMPIContext *context,
    const LMI_SELinuxServiceRef *self,
    KRef *Target,
    const KUint16 *Action,
    const KBoolean *Recursively,
    KRef *Job,
    CMPIStatus *__status)
{
    KUint32 result = KUINT32_INIT;
    CMPIObjectPath *cop;
    LmiJob *job;
    GVariant *v;
    LMI_UnixFileRef lmi_ufr;
    const gchar *instance_id;

    if (!KHasValue(Target)) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, LMI_PARAM_ERROR_STR(Target));
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    if (!KHasValue(Action)) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, LMI_PARAM_ERROR_STR(Action));
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }
    if (Action->value > 1) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, "Action out of range");
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    /* Recursively defaults to False */

    job = jobmgr_new_job(LMI_TYPE_JOB);
    if (job == NULL) {
        KSetStatus2(cb, __status, ERR_FAILED, "Could not create job instance");
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }
    lmi_job_set_method_name(job, "RestoreLabels");

    *__status = LMI_UnixFileRef_InitFromObjectPath(&lmi_ufr, cb, Target->value);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }
    if (!lmi_ufr.LFName.chars) {
        KSetStatus2(cb, __status, ERR_INVALID_PARAMETER, "The Target misses LFName");
        KUint32_Set(&result, JOB_RC_INVALID_PARAMETER);
        return result;
    }

    v = g_variant_new_string(lmi_ufr.LFName.chars);
    lmi_job_set_in_param(job, "Path", v);
    v = g_variant_new_int16(Action->value);
    lmi_job_set_in_param(job, "Action", v);
    v = g_variant_new_boolean(Recursively->value);
    lmi_job_set_in_param(job, "Recursively", v);
    v = g_variant_new_byte(JOB_RESTORE_LABELS);
    lmi_job_set_in_param(job, PARAM_TYPE, v);
    instance_id = lmi_get_string_property_from_objectpath(Target->value, "InstanceID");
    v = g_variant_new_string(instance_id);
    lmi_job_set_in_param(job, PARAM_INSTANCE_ID, v);

    *__status = jobmgr_run_job(job);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    *__status = jobmgr_job_to_cim_op(job, &cop);
    if (__status->rc) {
        KUint32_Set(&result, JOB_RC_FAILED);
        return result;
    }

    KRef_SetObjectPath(Job, cop);
    KUint32_Set(&result, JOB_RC_STARTED);
    return result;
}

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_SELinuxService",
    "LMI_SELinuxService",
    "instance method")
