/*
 * Copyright (C) 2014 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Jan Synacek <jsynacek@redhat.com>
 *
 */

#include <sys/file.h>
#include "selinux.h"

const ConfigEntry *provider_config_defaults = NULL;
const char *provider_name = "selinux";

void init_selinux_service(LMI_SELinuxService *service,
                          const CMPIBroker *b,
                          const CMPIContext *ctx,
                          const char *ns)
{
    LMI_SELinuxService_Init(service, b, ns);
    LMI_SELinuxService_Set_CreationClassName(service, LMI_SELinuxService_ClassName);
    LMI_SELinuxService_Set_Name(service, LMI_SELinuxService_ClassName);
    LMI_SELinuxService_Set_SystemCreationClassName(service, lmi_get_system_creation_class_name());
    LMI_SELinuxService_Set_SystemName(service, lmi_get_system_name_safe(ctx));
    LMI_SELinuxService_Set_InstanceID(service, LMI_ORGID ":" LMI_SELinuxService_ClassName);
}

CMPIStatus selinux_provider_init(const char *class,
                                 gboolean is_indication_provider,
                                 const ConfigEntry *provider_config_defaults,
                                 const CMPIBroker *cb,
                                 const CMPIContext *ctx)
{
    lmi_init(provider_name, cb, ctx, provider_config_defaults);
    lmi_debug("==> selinux_provider_init (class=%s)", class);

    jobmgr_register_job_type(LMI_TYPE_JOB, LMI_SELinuxJob_ClassName,
                             NULL, /* don't need convert_func */
                             job_make_params_callback,
                             METHOD_RESULT_VALUE_TYPE_UINT32,
                             job_process_callback,
                             FALSE,
                             FALSE);

    return jobmgr_init(cb, ctx, "SELinux", class, is_indication_provider, FALSE, NULL, 0);
}

FILE *open_selinux_config(const char *mode, off_t *size)
{
    const gchar *conf_path = "/etc/selinux/config";
    GStatBuf sb;
    FILE *f;

    f = fopen(conf_path, mode);
    if (f == NULL) {
        lmi_error("Can't open SELinux config file: %s (%s)", conf_path, mode);
        return NULL;
    }
    if (flock(fileno(f), LOCK_EX) < 0) {
        lmi_error("Can't acquire lock on '%s'", conf_path);
        fclose(f);
        return NULL;
    }

    if (g_stat(conf_path, &sb) < 0) {
        lmi_error("Stat on '%s' failed", conf_path);
        fclose(f);
        return NULL;
    }
    if (size)
        *size = sb.st_size;

    return f;
}

static gint set_selinux_default_state(gint newstate)
{
    gchar *content;
    off_t size = 0;
    size_t content_len = 0;
    gchar *line = NULL;
    size_t line_len = BUFLEN;
    ssize_t read;
    GRegex *re;
    GMatchInfo *mi;
    FILE *f;

    lmi_debug("==> set_selinux_default_state (newstate=%d)", newstate);

    /* Read the original config line by line, find the relevant config option
     * and if it's there, change it.
     */
    f = open_selinux_config("r", &size);
    if (f == NULL)
        return -1;

    /* add some buffer space */
    size += strlen("permissive");

    content = g_malloc0(size + 1);
    if (content == NULL) {
        lmi_error("Memory allocation failed");
        return -1;
    }

    re = g_regex_new("^\\s*SELINUX\\s*=", 0, 0, NULL);
    while((read = getline (&line, &line_len, f)) != -1) {
        if (g_regex_match(re, line, 0, &mi)) {
            read = snprintf(line, BUFLEN - 1, "SELINUX=%s\n", (newstate == 0) ? "disabled" :
                                                              (newstate == 1) ? "permissive" :
                                                              (newstate == 2) ? "enforcing" :
                                                                                "unknown");
            lmi_debug("SELINUX default state changed to %s", line);
        }
        g_strlcat(content, line, size);
        content_len += read;
    }

    g_match_info_free(mi);
    g_regex_unref(re);
    g_free(line);
    fclose(f);

    /* write new state */
    f = open_selinux_config("w", NULL);
    if (f == NULL) {
        free(content);
        return -1;
    }
    fwrite(content, 1, content_len, f);

    free(content);
    fclose(f);
    lmi_debug("<== set_selinux_default_state");
    return 0;
}

/* Callback: LMI_SELinuxService.SetSELinuxState() */
gint setselinuxstate(LmiJob *job)
{
    GVariant *v;
    gint rc;
    gint newstate;
    gboolean makedefault;

    lmi_debug("==> setselinuxstate");

    v = lmi_job_get_in_param(job, "NewState");
    newstate = g_variant_get_int16(v);
    v = lmi_job_get_in_param(job, "MakeDefault");
    makedefault = g_variant_get_boolean(v);

    lmi_debug("LMI_SELinuxService.SetSELinuxState: NewState: %d, MakeDefault: %d", newstate, makedefault);

    if (makedefault) {
        rc = set_selinux_default_state(newstate);
        if (rc < 0)
            lmi_job_finish_exception(job, LMI_JOB_STATUS_CODE_ENUM_FAILED, "Could not set default SELinux state");
    } else {
        rc = security_setenforce(newstate - 1);
        if (rc < 0)
            lmi_job_finish_exception(job, LMI_JOB_STATUS_CODE_ENUM_FAILED, "Could not set SELinux state");
    }

    lmi_debug("<== setselinuxstate (rc=%d)", rc);
    return rc;
}

/* Callback: LMI_SELinuxService.SetFileLabel() */
gint setfilelabel(LmiJob *job)
{
    GVariant *v;
    const gchar *path;
    const gchar *label;
    gint rc;

    lmi_debug("==> setfilelabel");

    v = lmi_job_get_in_param(job, "Path");
    path = g_variant_get_string(v, NULL);
    v = lmi_job_get_in_param(job, "Label");
    label = g_variant_get_string(v, NULL);

    lmi_debug("LMI_SELinuxService.SetFileLabel: Target: %s, Label: %s", path, label);

    security_context_t con;
    con = g_strdup(label);
    rc = setfilecon(path, con);
    freecon(con);

    if (rc < 0)
        lmi_job_finish_exception(job, LMI_JOB_STATUS_CODE_ENUM_FAILED,
                                 "Could not set file label: %s -> %s", path, label);

    lmi_debug("<== setfilelabel (rc=%d)", rc);
    return rc;
}

/* Callback: LMI_SELinuxService.SetPortLabel() */
gint setportlabel(LmiJob *job)
{
    gint rc = -1;
    GVariant *v;
    const gchar *name;
    gint type, range_low, range_high;
    semanage_handle_t *hnd = NULL;
    semanage_port_t *port = NULL;
    semanage_port_key_t *pkey = NULL;
    semanage_context_t *ctx = NULL;

    lmi_debug("==> setportlabel");

    hnd = semanage_handle_create();
    if (hnd == NULL) {
        lmi_error("semanage_handle_create failed");
        return rc;
    }

    v = lmi_job_get_in_param(job, "Name");
    name = g_variant_get_string(v, NULL);
    v = lmi_job_get_in_param(job, "Type");
    type = g_variant_get_int32(v);
    v = lmi_job_get_in_param(job, "RangeLow");
    range_low = g_variant_get_int32(v);
    v = lmi_job_get_in_param(job, "RangeHigh");
    range_high = g_variant_get_int32(v);

    lmi_debug("LMI_SELinuxService.SetPortLabel: Name: %s, Type: %d, RangeLow: %d, RangeHigh: %d",
              name, type, range_low, range_high);

    rc = semanage_connect(hnd);
    if (rc < 0) {
        lmi_error("semanage_connect failed");
        goto err;
    }

    rc = semanage_port_key_create(hnd, range_low, range_high, type, &pkey);
    if (rc < 0) {
        lmi_error("semanage_port_key_create failed");
        goto err;
    }

    rc = semanage_port_query(hnd, pkey, &port);
    if (rc < 0) {
        lmi_error("semanage_port_query failed");
        goto err;
    }
    if (port == NULL) {
        lmi_error("%s port %d-%d does not exist", PORT_ENUM_TO_STR(type), range_low, range_high);
        rc = -1;
        goto err;
    }

    ctx = semanage_port_get_con(port);
    if (ctx == NULL) {
        lmi_error("Port context is not valid");
        rc = -1;
        goto err;
    }

    lmi_debug("Found '%s' label on that port", semanage_context_get_type(ctx));

    rc = semanage_context_set_type(hnd, ctx, name);
    if (rc < 0) {
        lmi_error("semanage_context_set_type failed");
        goto err;
    }

    rc = semanage_port_set_con(hnd, port, ctx);
    if (rc < 0) {
        lmi_error("semanage_port_set_con failed");
        goto err;
    }

    /* modification */
    rc = semanage_begin_transaction(hnd);
    if (rc < 0) {
        lmi_error("semanage_begin_transaction failed");
        goto err;
    }
    rc = semanage_port_modify_local(hnd, pkey, port);
    if (rc < 0) {
        lmi_error("semanage_port_modify_local failed");
        goto err;
    }
    rc = semanage_commit(hnd);
    if (rc < 0)
        lmi_error("semanage_commit failed");
err:
    semanage_port_key_free(pkey);
    semanage_port_free(port);
    semanage_disconnect(hnd);
    semanage_handle_destroy(hnd);
    if (rc < 0)
        lmi_job_finish_exception(job, LMI_JOB_STATUS_CODE_ENUM_FAILED,
                                 "Could not label %s port %d-%d with %s",
                                 PORT_ENUM_TO_STR(type), range_low, range_high, name);

    lmi_debug("<== setportlabel (rc=%d)", rc);
    return rc;
}

/* Callback: LMI_SELinuxService.SetBoolean() */
gint setboolean(LmiJob *job)
{
    GVariant *v;
    gchar *boolname;
    gboolean value;
    gboolean makedefault;
    int exists = 0;
    semanage_handle_t *hnd = NULL;
    semanage_bool_key_t *bk = NULL;
    semanage_bool_t *b = NULL;
    gint rc = -1;

    lmi_debug("==> setboolean");

    v = lmi_job_get_in_param(job, "BoolName");
    boolname = (gchar *) g_variant_get_string(v, NULL);
    v = lmi_job_get_in_param(job, "Value");
    value = g_variant_get_boolean(v);
    v = lmi_job_get_in_param(job, "MakeDefault");
    makedefault = g_variant_get_boolean(v);

    lmi_debug("LMI_SELinuxService.SetBoolean: Target: %s, Value: %d, MakeDefault: %d",
              boolname, value, makedefault);

    /* set permanent boolean value */
    /* For some reason, security_set_boolean_list() called with permanent = 1 doesn't work,
     * so semanage has to be used. */
    if (makedefault) {
        hnd = semanage_handle_create();
        if (!hnd) {
            lmi_error("semanage_handle_create failed");
            return rc;
        }

        rc = semanage_connect(hnd);
        if (rc < 0) {
            lmi_error("semanage_connect failed");
            goto err;
        }

        /* gets a semanage lock internally */
        rc = semanage_begin_transaction(hnd);
        if (rc < 0) {
            lmi_error("semanage_begin_transaction failed");
            goto err;
        }

        rc = semanage_bool_create(hnd, &b);
        if (rc < 0) {
            lmi_error("semanage_bool_create failed");
            goto err;
        }

        rc = semanage_bool_set_name(hnd, b, boolname);
        if (rc < 0) {
            lmi_error("semanage_bool_set_name failed");
            goto err;
        }
        semanage_bool_set_value(b, value);

        rc = semanage_bool_key_extract(hnd, b, &bk);
        if (rc < 0) {
            lmi_error("semanage_bool_key_extract failed");
            goto err;
        }

        rc = semanage_bool_exists(hnd, bk, &exists);
        if (rc < 0) {
            lmi_error("semanage_bool_exists failed");
            goto err;
        }
        if (!exists) {
            rc = semanage_bool_exists_local(hnd, bk, &exists);
            if (rc < 0) {
                lmi_error("semanage_bool_exists_local failed");
                goto err;
            }
            if (!exists) {
                lmi_error("Boolean %s does not exist", boolname);
                goto err;
            }
        }

        rc = semanage_bool_modify_local(hnd, bk, b);
        if (rc < 0) {
            lmi_error("semanage_bool_modify_local failed");
            goto err;
        }

        rc = semanage_bool_set_active(hnd, bk, b);
        if (rc < 0) {
            lmi_error("semanage_bool_set_active failed");
            goto err;
        }

        semanage_set_reload(hnd, 1);

        rc = semanage_commit(hnd);
        if (rc < 0) {
            lmi_error("semanage_commit failed");
            goto err;
        }

    } else { /* makedefault == false */
        /* set current boolean value */
        SELboolean boollist = {boolname, value};
        rc = security_set_boolean_list(1, &boollist, 0);
        if (rc < 0) {
            lmi_debug("Setting boolean %s to %d failed", boolname, value);
            return rc;
        }
    }

err:
    /* semanage*_free functions handle NULLs fine */
    semanage_bool_key_free(bk);
    semanage_bool_free(b);
    semanage_disconnect(hnd);
    semanage_handle_destroy(hnd);
    if (rc < 0)
        lmi_job_finish_exception(job, LMI_JOB_STATUS_CODE_ENUM_FAILED, "Could not set boolean %s", boolname);

    lmi_debug("<== setboolean (rc=%d)", rc);
    return rc;
}

void job_process_callback(LmiJob *job, GCancellable *cancellable)
{
    guchar _type;
    gint rc = 0;
    const gchar *err;

    lmi_debug("==> job_process_callback");

    if (cancellable && g_cancellable_is_cancelled(cancellable))
        return;

    _type = g_variant_get_byte(lmi_job_get_in_param(job, PARAM_TYPE));
    switch (_type) {
    case JOB_SET_SELINUX_STATE:
        rc = setselinuxstate(job);
        if (rc < 0)
            err = "Setting SELinux state failed";
        break;
    case JOB_RESTORE_LABELS:
        rc = restore_labels(job);
        if (rc < 0)
            err = "Restoring SELinux labels failed";
        break;
    case JOB_SET_FILE_LABEL:
        rc = setfilelabel(job);
        if (rc < 0)
            err = "Setting file label failed";
        break;
    case JOB_SET_PORT_LABEL:
        rc = setportlabel(job);
        if (rc < 0)
            err = "Setting port label failed";
        break;
    case JOB_SET_BOOLEAN:
        rc = setboolean(job);
        if (rc < 0)
            err = "Setting SELinux boolean failed";
        break;
    default:
        rc = -1;
        lmi_job_finish_exception(job, LMI_JOB_STATUS_CODE_ENUM_FAILED, "Fatal: Unknown internal job type");
    }

    if (!lmi_job_is_finished(job)) {
        if (rc < 0)
            lmi_job_finish_exception(job, LMI_JOB_STATUS_CODE_ENUM_FAILED, err);
        else
            lmi_job_finish_ok(job, JOB_RC_COMPLETED);
    }
    lmi_debug("<== job_process_callback");
}

CMPIStatus job_make_params_callback(const CMPIBroker *cb,
                                    const CMPIContext *ctx,
                                    const LmiJob *job,
                                    gboolean include_input,
                                    gboolean include_output,
                                    CMPIInstance *instance)
{
    CMPIStatus st = {CMPI_RC_OK, NULL};
    CMPIValue val;
    CMPIObjectPath *o;
    LMI_UnixFileRef lmi_ufr;
    const char *ns;
    struct restore_labels_data *rldata;
    guchar type;

    lmi_debug("==> job_make_params_callback");

    if (!include_output)
        return st;

    type = g_variant_get_byte(lmi_job_get_in_param(job, PARAM_TYPE));

    if (type == JOB_RESTORE_LABELS) {
        rldata = (struct restore_labels_data *) lmi_job_get_data(job);
        o = CMGetObjectPath(instance, NULL);
        val.string = CMGetNameSpace(o, NULL);
        ns = val.string->ft->getCharPtr(val.string, NULL);
        LMI_UnixFileRef_Init(&lmi_ufr, cb, ns);
        LMI_UnixFileRef_Set_LFName(&lmi_ufr, rldata->path);
        LMI_UnixFileRef_Set_CSCreationClassName(&lmi_ufr, lmi_get_system_creation_class_name());
        LMI_UnixFileRef_Set_CSName(&lmi_ufr, lmi_get_system_name_safe(ctx));

        o = LMI_UnixFileRef_ToObjectPath(&lmi_ufr, &st);
        if (st.rc)
            return st;
        val.ref = o;
        st = CMSetProperty(instance, "Target", &val, CMPI_ref);
        if (st.rc)
            return st;
    }

    val.uint32 = JOB_RC_COMPLETED;
    st = CMSetProperty(instance, "__ReturnValue", &val, CMPI_uint32);
    lmi_debug("<== job_make_params_callback");
    return st;
}

/**
 * Handler for actual SELinux context restoring.
 * Returns rc < 0 on error.
 */
static gint restore_label_handler(const gchar *path, struct selabel_handle *hnd, void *data)
{
    security_context_t c_current, c_expected;
    struct restore_labels_data *rldata = (struct restore_labels_data *)data;
    gint rc, mismatch;

    lmi_debug("==> restore_label_handler (file=%s)", path);

    rc = lgetfilecon(path, &c_current);
    if (rc < 0) {
        lmi_error("lgetfilecon() failed on '%s'\n", path);
        return rc;
    }

    rc = selabel_lookup(hnd, &c_expected, path, 0);
    if (rc < 0) {
        lmi_error("selabel_lookup() failed on '%s'\n", path);
        freecon(c_current);
        return rc;
    }

    mismatch = strcmp(c_current, c_expected);
    if (mismatch)
        lmi_info("Context mismatch on '%s': current:%s, expected:%s", path, c_current, c_expected);

    if (mismatch && rldata->action == ACTION_RESTORE) {
        lmi_info("Restoring context on '%s'", path);
        rc = selinux_lsetfilecon_default(path);
        if (rc)
            lmi_error("Could not restore file context on '%s'", path);
    }
    /* do nothing when action == restore, files should be available (reported)
     * via LMI_AffectedSELinuxJobElement */

    freecon(c_expected);
    freecon(c_current);
    lmi_debug("<== restore_label_handler (rc=%d)", rc);
    return rc;
}

/**
 * Restore SELinux context on files in the specified directory. If specified,
 * traverse the directory structure recursively.
 *
 * A hash table with already traversed paths is maintained to avoid being stuck in
 * cycles.
 */
static gint restore_label_pvt(struct restore_labels_data *rldata)
{
    GError *gerror = NULL;
    GDir *gdir = NULL;
    const gchar *gde;
    gchar *filepath = NULL;
    gchar *path = rldata->path;
    GQueue *paths;
    GStatBuf sb;
    struct selabel_handle *label_hnd;
    gint rc;

    lmi_debug("==> restore_label_pvt (path=%s)", path);

    label_hnd = selabel_open(SELABEL_CTX_FILE, NULL, 0);
    if (label_hnd == NULL) {
        lmi_error("selabel_open() failed\n");
        return -1;
    }

    rc = g_stat(path, &sb);
    if (rc < 0) {
        lmi_error("Stat on '%s' failed", path);
        goto out;
    }

    if (!S_ISDIR(sb.st_mode))
        goto out;

    if (!g_hash_table_contains(rldata->paths_htable, path)) {
        restore_label_handler(path, label_hnd, (void *)rldata);
    }
    g_hash_table_add(rldata->paths_htable, g_strdup(path));

    paths = g_queue_new();
    g_queue_push_head(paths, g_strdup(path));

    while (!g_queue_is_empty(paths)) {
        path = (gchar *)g_queue_pop_head(paths);

        gdir = g_dir_open(path, 0, &gerror);
        if (gerror) {
            lmi_error("Opening '%s' failed: %s", path, gerror->message);
            continue;
        }

        while ((gde = g_dir_read_name(gdir))) {
            gchar *actualpath;

            filepath = g_build_filename(path, gde, NULL);
            g_assert(filepath);
            gerror = NULL;

            /* ignore rc, the handler logs any errors */
            restore_label_handler(filepath, label_hnd, (void *)rldata);

            rc = g_stat(filepath, &sb);
            if (rc < 0) {
                g_free(filepath);
                continue;
            }

            if (g_hash_table_contains(rldata->paths_htable, filepath)) {
                /* already exists, throw warning and continue; */
                lmi_debug("Already been there, skipping: %s\n", filepath);
                g_free(filepath);
                continue;
            }
            g_hash_table_add(rldata->paths_htable, g_strdup(filepath));

            /* normalize file path in case of symlink */
            actualpath = realpath(filepath, NULL);
            /* some dynamic files don't have to exist at this point */
            if (actualpath == NULL && errno == ENOENT) {
                lmi_info("File disappeared: %s\n", filepath);
                g_free(filepath);
                continue;
            }
            g_free(filepath);
            filepath = actualpath;

            if (rldata->recursively && S_ISDIR(sb.st_mode))
                g_queue_push_head(paths, filepath);
        }
        g_dir_close(gdir);
    }
    g_queue_free_full(paths, g_free);
out:
    selabel_close(label_hnd);
    lmi_debug("<== restore_label_pvt (rc=%d)", rc);
    return rc;
}

/* Callback: LMI_SELinuxService.RestoreLabels() */
static void job_data_free(gpointer data)
{
    struct restore_labels_data *rldata = (struct restore_labels_data *)data;
    g_free(rldata->path);
    g_hash_table_destroy(rldata->paths_htable);
}

gint restore_labels(LmiJob *job)
{
    GVariant *v;
    gchar *path;
    gint action;
    gboolean recursively;
    gint rc;
    GHashTable *paths_htable;
    struct restore_labels_data *rldata;

    lmi_debug("==> restore_labels");

    v = lmi_job_get_in_param(job, "Path");
    path = (gchar *) g_variant_get_string(v, NULL);
    v = lmi_job_get_in_param(job, "Action");
    action = g_variant_get_int16(v);
    v = lmi_job_get_in_param(job, "Recursively");
    recursively = g_variant_get_boolean(v);

    lmi_debug("LMI_SELinuxService.RestoreLabels: Target: %s, Action: %d, Recursively: %d",
              path, action, recursively);

    /* This hash table is used in such way that keys == values.
     * Only one g_free is necessary to free both. */
    paths_htable = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);

    rldata = g_malloc(sizeof(struct restore_labels_data));
    if (!rldata) {
        lmi_error("Memory allocation failed");
        return -1;
    }
    rldata->action = action;
    rldata->recursively = recursively;
    rldata->path = path;
    rldata->job = job;
    rldata->paths_htable = paths_htable;
    lmi_job_set_data(job, (struct restore_labels_data *) rldata, job_data_free);

    rc = restore_label_pvt(rldata);

    lmi_debug("<== restore_labels (rc=%d)", rc);
    return rc;
}

/**
 * Return list of InstanceIDs of all LMI_SELinuxPort instances or NULL on
 * error. The caller must free the list using g_slist_free_full(list, free).
 */
GSList *lmi_get_selinux_port_names()
{
    GSList *list = NULL;
    int rc;
    semanage_handle_t *manage_hnd = NULL;
    semanage_port_t **ports = NULL;
    unsigned int n_ports;
    unsigned int i = 0;
    GHashTable *ports_hash = NULL;
    gboolean connected = FALSE;

    manage_hnd = semanage_handle_create();
    if (manage_hnd == NULL) {
        return NULL;
    }

    rc = semanage_connect(manage_hnd);
    if (rc < 0)
        goto err;
    connected = TRUE;

    rc = semanage_port_list(manage_hnd, &ports, &n_ports);
    if (rc < 0)
        goto err;

    ports_hash = g_hash_table_new(g_str_hash, g_str_equal);
    for (; i < n_ports; i++) {
        semanage_context_t *ctx = semanage_port_get_con(ports[i]);
        const char *p_type = semanage_context_get_type(ctx);
        const char *p_proto = PORT_PROTO(ports[i]);

        char *instance_id;
        rc = asprintf(&instance_id,
                      LMI_ORGID ":" LMI_SELinuxPort_ClassName ":%s:%s", p_proto, p_type);
        if (rc < 0)
            goto err;

        /* Lookup the value first, we do not want duplicates in the returned list.*/
        if (!g_hash_table_contains(ports_hash, instance_id)) {
            g_hash_table_insert(ports_hash, instance_id, NULL);
            list = g_slist_prepend(list, instance_id);
        } else {
            free(instance_id);
        }
    }

err:
    if (ports) {
        for (i=0; i<n_ports; i++)
            semanage_port_free(ports[i]);
        free(ports);
    }
    if (connected)
        semanage_disconnect(manage_hnd);
    semanage_handle_destroy(manage_hnd);
    if (ports_hash)
        g_hash_table_destroy(ports_hash);
    if (rc < 0) {
        g_slist_free_full(list, free);
        list = NULL;
    }
    return list;
}

/**
 * Return list of InstanceIDs of all LMI_SELinuxBoolean instances or NULL on
 * error. The caller must free the list using g_slist_free_full(list, free).
 */
GSList *lmi_get_selinux_boolean_names()
{
    GSList *list = NULL;
    int rc;
    int len;
    char **booleans;

    rc = security_get_boolean_names(&booleans, &len);
    if (rc < 0) {
        return NULL;
    }

    for (int i = 0; i < len; i++) {
        char *instance_id;
        rc = asprintf(&instance_id,
                      LMI_ORGID ":" LMI_SELinuxBoolean_ClassName ":%s", booleans[i]);
        if (rc < 0)
            goto err;
        list = g_slist_prepend(list, instance_id);
    }

err:
    for (int i=0; i< len; i++)
        free(booleans[i]);
    free(booleans);

    if (rc < 0) {
        g_slist_free_full(list, free);
        list = NULL;
    }
    return list;
}
