/*
 * Copyright (C) 2014 Red Hat, Inc.  All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Authors: Jan Synacek <jsynacek@redhat.com>
 *
 */

#ifndef _SELINUX_PROVIDER_H
#define _SELINUX_PROVIDER_H

#include <selinux/label.h>
#include <selinux/selinux.h>
#include <semanage/semanage.h>
#include <glib.h>
#include <glib/gstdio.h>	/* GStatBuf */

#include "LMI_SELinuxService.h"
#include "openlmi.h"
#include "job_manager.h"

enum {
    JOB_SET_SELINUX_STATE = 0x10,
    JOB_RESTORE_LABELS,
    JOB_SET_FILE_LABEL,
    JOB_SET_PORT_LABEL,
    JOB_SET_BOOLEAN,
    JOB_NUM
};

enum {
    ACTION_REPORT,
    ACTION_RESTORE,
};

#define JOB_RC_COMPLETED            0
#define JOB_RC_FAILED               4
#define JOB_RC_INVALID_PARAMETER    5
#define JOB_RC_STARTED           4096

/* Special job parameter that specifies the type of action that the job is supposed to do.
 * Used when job "inheritance" is deemed to be overkill. */
#define PARAM_TYPE "_Type"

/* Special job parameter specifying the entire InstanceID of the Target method parameter.
 * Used to create instance references in AffectedJobElement. */
#define PARAM_INSTANCE_ID "_TargetInstanceID"

#define PORT_STR_TO_ENUM(pstr)  ((strcmp(pstr, "TCP") == 0) ? SEMANAGE_PROTO_TCP : SEMANAGE_PROTO_UDP)
#define PORT_ENUM_TO_STR(e)     ((e == SEMANAGE_PROTO_TCP) ? "TCP" : "UDP")
#define PORT_PROTO(port)        (PORT_ENUM_TO_STR(semanage_port_get_proto(port)))
#define PORT_STR_IS_VALID(pstr) ((strcmp(pstr, "TCP") == 0) || (strcmp(pstr, "UDP") == 0))

struct cmpi_data {
    const CMPIResult *cr;
    const char *namespace;
};

struct restore_labels_data {
    gint action;
    gboolean recursively;
    gchar *path;
    LmiJob *job;
    GHashTable *paths_htable;
};

struct mislabeled_file {
    security_context_t sec_ctx_current;
    security_context_t sec_ctx_expected;
    gchar *path;
};

const ConfigEntry *provider_config_defaults;
const char *provider_name;

/**
 * Safely open the SELinux configuration file.
 *
 * Acquires a file lock on the file.
 *
 * @param mode Mode in which to open the file.
 * @param size If not NULL, size is set to the file's actual size in bytes.
 */
FILE *open_selinux_config(const char *mode, off_t *size);

/**
 * Initialize an instance of LMI_SELinuxService.
 *
 * @param service Instance of LMI_SELinuxService.
 * @param b CMPIBroker.
 * @param ctx CMPIContext.
 * @param ns Namespace.
 */
void init_selinux_service(LMI_SELinuxService *service,
                          const CMPIBroker *b,
                          const CMPIContext *ctx,
                          const char *ns);

/* job specific */
CMPIStatus selinux_provider_init(const char *class,
                                 gboolean is_indication_provider,
                                 const ConfigEntry *provider_config_defaults,
                                 const CMPIBroker *cb,
                                 const CMPIContext *ctx);
CMPIStatus job_make_params_callback(const CMPIBroker *cb,
                                    const CMPIContext *ctx,
                                    const LmiJob *job,
                                    gboolean include_input,
                                    gboolean include_output,
                                    CMPIInstance *instance);
void job_process_callback(LmiJob *job, GCancellable *cancellable);

gint setselinuxstate(LmiJob *job);
gint restore_labels(LmiJob *job);
gint setfilelabel(LmiJob *job);
gint setboolean(LmiJob *job);

/**
 * Return list of InstanceIDs of all LMI_SELinuxBoolean instances or NULL on
 * error. The caller must free the list using g_slist_free_full(list, free).
 */
GSList *lmi_get_selinux_boolean_names();
/**
 * Return list of InstanceIDs of all LMI_SELinuxPort instances or NULL on
 * error. The caller must free the list using g_slist_free_full(list, free).
 */
GSList *lmi_get_selinux_port_names();

#endif
