#include <konkret/konkret.h>
#include "LMI_SoftwareInstallationServiceAffectsElement.h"
#include "LMI_SoftwareIdentity.h"
#include "sw-utils.h"

#define LMI_AFFECTING_ELEMENT "AffectingElement"
#define LMI_AFFECTED_ELEMENT  "AffectedElement"

static const CMPIBroker* _cb;

static void LMI_SoftwareInstallationServiceAffectsElementInitialize(const CMPIContext *ctx)
{
    software_init(LMI_SoftwareInstallationServiceAffectsElement_ClassName,
            _cb, ctx, FALSE, provider_config_defaults);
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementCleanup(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    return software_cleanup(LMI_SoftwareInstallationServiceAffectsElement_ClassName);
}

static CMPIStatus k_return_sisaep(PkPackage *pk_pkg,
        const LMI_SoftwareInstallationServiceRef *sis, const CMPIResult *cr,
        const char *ns, const short names)
{
    CMPIStatus st = {CMPI_RC_OK, NULL};
    SwPackage sw_pkg;
    char elem_name[BUFLEN] = "", instance_id[BUFLEN] = "";

    init_sw_package(&sw_pkg);

    if (create_sw_package_from_pk_pkg(pk_pkg, &sw_pkg) != 0) {
        KSetStatus(&st, ERR_FAILED);
        goto done;
    }

    sw_pkg_get_element_name(&sw_pkg, elem_name, BUFLEN);
    clean_sw_package(&sw_pkg);

    create_instance_id(LMI_SoftwareIdentity_ClassName, elem_name, instance_id,
            BUFLEN);

    LMI_SoftwareIdentityRef si;
    LMI_SoftwareIdentityRef_Init(&si, _cb, ns);
    LMI_SoftwareIdentityRef_Set_InstanceID(&si, instance_id);

    LMI_SoftwareInstallationServiceAffectsElement w;
    LMI_SoftwareInstallationServiceAffectsElement_Init(&w, _cb, ns);
    LMI_SoftwareInstallationServiceAffectsElement_Set_AffectingElement(&w, sis);
    LMI_SoftwareInstallationServiceAffectsElement_Set_AffectedElement(&w,
            (const CIM_ManagedElementRef *) &si);

    if (names) {
        KReturnObjectPath(cr, w);
    } else {
        KReturnInstance(cr, w);
    }

done:
    return st;
}

/**
 * Additional arguments container for `traverse_avail_pkg_repos()`.
 */
struct TraverseAvailPkgReposData {
    const CMPIBroker *cb;
    const CMPIContext *ctx;
    const gchar *ns;
    const CMPIResult *cr;
    /* whether to yield instance names or full instances */
    short names;
    LMI_SoftwareInstallationServiceRef *sis;
};

/**
 * Traversing function for a hash/tree with `PkPackages *` being keys and
 * NULL-terminated array of `PkRepoDetail *` being their associated values.
 *
 * It yields instance or instance name of `LMI_SoftwareInstallationServiceAffectsElement` for
 * given pair key/value.
 *
 * @param data Pointer to `TraverseAvailPkgReposData`.
 */
static gboolean traverse_avail_pkg_repos(gpointer key,
                                         gpointer value,
                                         gpointer data)
{
    struct TraverseAvailPkgReposData *taprd = data;
    PkPackage *pk_pkg = PK_PACKAGE(key);
    CMPIStatus st;

    st = k_return_sisaep(pk_pkg, taprd->sis, taprd->cr, taprd->ns, taprd->names);

    return st.rc != CMPI_RC_OK;
}

/**
 * Traversing function for a hash/tree with `PkPackages *` being keys and
 * NULL-terminated array of `PkRepoDetail *` being their associated values.
 *
 * It yields instance name of `LMI_SoftwareIdentity` for
 * given pair key/value.
 *
 * @param data Pointer to `TraverseAvailPkgReposData`.
 */
static gboolean traverse_avail_pkg_repos_for_sw(gpointer key,
                                                gpointer value,
                                                gpointer data)
{
    struct TraverseAvailPkgReposData *taprd = data;
    PkPackage *pk_pkg = PK_PACKAGE(key);
    CMPIStatus st;

    st = k_return_sw_identity_op_from_pkg_id(pk_package_get_id(pk_pkg),
            taprd->cb, taprd->ctx, taprd->ns, taprd->cr);

    return st.rc != CMPI_RC_OK;
}

static CMPIStatus enum_instances(
    const CMPIContext *ctx,
    const CMPIResult *cr,
    const char *ns,
    const short names)
{
    CMPIStatus st;
    struct TraverseAvailPkgReposData taprd = {NULL, NULL, ns, cr, names, NULL};
    GTree *avail_pkg_repos = NULL;
    GPtrArray *array = NULL;
    CMPIObjectPath *op = NULL;
    char error_msg[BUFLEN] = "";

    st = get_software_installation_service_op(_cb, ctx, &op);
    if (st.rc)
        goto done;

    LMI_SoftwareInstallationServiceRef sis;
    LMI_SoftwareInstallationServiceRef_InitFromObjectPath(&sis, _cb, op);

    LMI_SoftwareInstallationServiceAffectsElement w;
    LMI_SoftwareInstallationServiceAffectsElement_Init(&w, _cb, ns);
    LMI_SoftwareInstallationServiceAffectsElement_Set_AffectingElement(&w, &sis);
    LMI_SoftwareInstallationServiceAffectsElement_SetObjectPath_AffectedElement(&w,
            lmi_get_computer_system_safe(ctx));

    if (names) {
        KReturnObjectPath(cr, w);
    } else {
        KReturnInstance(cr, w);
    }

    taprd.sis = &sis;

    get_pk_packages(0, &array, error_msg, BUFLEN);
    if (!array)
        goto done;

    /* Associate packages to repositories providing them.
     * `avail_pkg_repos` will contain just available packages. */
    get_repo_dets_for_pk_pkgs(array, NULL, &avail_pkg_repos, error_msg, BUFLEN);
    if (!avail_pkg_repos)
        goto done;

    /* Generate instances/instance names for obtained associations. */
    g_tree_foreach(avail_pkg_repos, traverse_avail_pkg_repos, &taprd);

done:
    if (avail_pkg_repos) {
        g_tree_unref(avail_pkg_repos);
    }
    if (array) {
        g_ptr_array_unref(array);
    }

    if (*error_msg) {
        KReturn2(_cb, ERR_FAILED, "%s", error_msg);
    }

    return st;
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementEnumInstanceNames(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    return enum_instances(cc, cr, KNameSpace(cop), 1);
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementEnumInstances(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    return enum_instances(cc, cr, KNameSpace(cop), 0);
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementGetInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char** properties)
{
    CMPIStatus st = {CMPI_RC_OK, NULL};
    SwPackage sw_pkg;
    gchar *repo_id = NULL;
    char error_msg[BUFLEN] = "", instance_id[BUFLEN] = "";

    init_sw_package(&sw_pkg);

    create_instance_id(LMI_SoftwareInstallationService_ClassName, NULL,
            instance_id, BUFLEN);

    LMI_SoftwareInstallationServiceAffectsElement w;
    LMI_SoftwareInstallationServiceAffectsElement_InitFromObjectPath(&w, _cb, cop);

    st = check_software_installation_service(_cb, cc, w.AffectingElement.value);
    if (st.rc != 0)
        goto done;

    if (CMClassPathIsA(_cb, w.AffectedElement.value,
                LMI_SoftwareIdentity_ClassName, &st)) {
        if (get_sw_pkg_from_sw_identity_op(w.AffectedElement.value, &sw_pkg) != 0) {
            KSetStatus(&st, ERR_NOT_FOUND);
            goto done;
        }

        get_repo_id_from_sw_pkg(&sw_pkg, &repo_id, error_msg, BUFLEN);
        if (!repo_id) {
            KSetStatus(&st, ERR_NOT_FOUND);
            goto done;
        }
    } else {
        if (!lmi_check_computer_system(cc, w.AffectedElement.value)) {
            KSetStatus(&st, ERR_NOT_FOUND);
            goto done;
        }
    }

    KReturnInstance(cr, w);

done:
    clean_sw_package(&sw_pkg);
    g_free(repo_id);

    if (*error_msg) {
        KReturn2(_cb, ERR_FAILED, "%s", error_msg);
    }

    return st;
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementCreateInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementModifyInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const CMPIInstance* ci,
    const char**properties)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementDeleteInstance(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementExecQuery(
    CMPIInstanceMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* lang,
    const char* query)
{
    CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementAssociationCleanup(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    CMPIBoolean term)
{
    CMReturn(CMPI_RC_OK);
}

static CMPIStatus associators(
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* resultClass,
    const char* role,
    const char* resultRole,
    const char** properties,
    const short names)
{
    CMPIStatus st = {CMPI_RC_OK, NULL};
    SwPackage sw_pkg;
    gchar *repo_id = NULL;
    GPtrArray *array = NULL;
    GTree *avail_pkg_repos = NULL;
    CMPIObjectPath *op = NULL;
    char error_msg[BUFLEN] = "";

    st = lmi_class_path_is_a(_cb, KNameSpace(cop),
            LMI_SoftwareInstallationServiceAffectsElement_ClassName, assocClass);
    lmi_return_if_class_check_not_ok(st);

    init_sw_package(&sw_pkg);

    if (CMClassPathIsA(_cb, cop, LMI_SoftwareInstallationService_ClassName, &st)) {
        /* got SoftwareInstallationService - AffectingElement;
         * where listing only associators names is supported */
        struct TraverseAvailPkgReposData taprd = {_cb, cc, KNameSpace(cop), cr,
                names, NULL};

        if (!names) {
            lmi_error("Just associator names can be enumerated for %s!",
                    LMI_SoftwareInstallationService_ClassName);
            CMReturn(CMPI_RC_ERR_NOT_SUPPORTED);
        }

        if (resultClass && resultClass[0]) {
            st = lmi_class_path_is_a(_cb, KNameSpace(cop),
                    resultClass, LMI_SoftwareIdentity_ClassName);
            if (st.rc != 0)
                st = lmi_class_path_is_a(_cb, KNameSpace(cop),
                        resultClass, CIM_ComputerSystem_ClassName);
            lmi_return_if_class_check_not_ok(st);
        }

        if (role && strcasecmp(role, LMI_AFFECTING_ELEMENT) != 0) {
            lmi_debug("Unexpected role \"%s\".", role);
            goto done;
        }
        if (resultRole && strcasecmp(resultRole, LMI_AFFECTED_ELEMENT) != 0) {
            lmi_debug("Unexpected resultRole \"%s\".", resultRole);
            goto done;
        }

        if (resultClass == NULL || lmi_class_path_is_a(_cb, KNameSpace(cop),
                    resultClass, LMI_SoftwareIdentity_ClassName).rc == 0) {

            get_pk_packages(0, &array, error_msg, BUFLEN);
            if (!array) {
                goto done;
            }

            /* Associate packages to repositories providing them.
             * `avail_pkg_repos` will contain just available packages. */
            get_repo_dets_for_pk_pkgs(array, NULL, &avail_pkg_repos, error_msg, BUFLEN);
            if (!avail_pkg_repos) {
                goto done;
            }

            /* Generate instance names for obtained associations. */
            g_tree_foreach(avail_pkg_repos, traverse_avail_pkg_repos_for_sw, &taprd);
        }

        if (resultClass == NULL || lmi_class_path_is_a(_cb, KNameSpace(cop),
                    resultClass, CIM_ComputerSystem_ClassName).rc == 0) {
            CMReturnObjectPath(cr, lmi_get_computer_system_safe(cc));
        }

    } else {
        /* got AffectedElement */
        if (role && strcasecmp(role, LMI_AFFECTED_ELEMENT) != 0) {
            lmi_debug("Unexpected role \"%s\".", role);
            goto done;
        }
        if (resultRole && strcasecmp(resultRole, LMI_AFFECTING_ELEMENT) != 0) {
            lmi_debug("Unexpected resultRole \"%s\".", resultRole);
            goto done;
        }

        st = lmi_class_path_is_a(_cb, KNameSpace(cop),
                LMI_SoftwareInstallationService_ClassName, resultClass);
        lmi_return_if_class_check_not_ok(st);

        if (CMClassPathIsA(_cb, cop, CIM_ComputerSystem_ClassName, &st)) {
            /* got ComputerSystem */
            if (!lmi_check_computer_system(cc, cop)) {
                KSetStatus(&st, ERR_NOT_FOUND);
                goto done;
            }
        } else if (CMClassPathIsA(_cb, cop, LMI_SoftwareIdentity_ClassName, &st)) {
            /* got SoftwareIdentitty */
            if (get_sw_pkg_from_sw_identity_op(cop, &sw_pkg) != 0) {
                lmi_debug("Failed to process object path of %s.",
                        LMI_SoftwareIdentity_ClassName);
                goto done;
            }
            get_repo_id_from_sw_pkg(&sw_pkg, &repo_id, error_msg, BUFLEN);
            if (!repo_id)
                goto done;
        } else {
            lmi_debug("Unknown association class \"%s\".",
                    CMGetCharsPtr(CMGetClassName(cop, &st), NULL));
            goto done;
        }

        st = get_software_installation_service_op(_cb, cc, &op);
        if (st.rc != 0)
            goto done;
        if (names) {
            CMReturnObjectPath(cr, op);
        } else {
            CMPIInstance *ci = _cb->bft->getInstance(_cb, cc, op, properties, &st);
            CMReturnInstance(cr, ci);
        }
    }

done:
    clean_sw_package(&sw_pkg);
    g_clear_pointer(&avail_pkg_repos, g_tree_unref);
    g_clear_pointer(&array, g_ptr_array_unref);
    g_free(repo_id);

    if (*error_msg) {
        lmi_error("Failed to list associators of \"%s\": %s",
                CMGetCharsPtr(CMGetClassName(cop, NULL), NULL), error_msg);
        CMSetStatusWithChars(_cb, &st,
                st.rc == CMPI_RC_OK ? CMPI_RC_ERR_FAILED : st.rc, error_msg);
    }

    return st;
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementAssociators(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* resultClass,
    const char* role,
    const char* resultRole,
    const char** properties)
{
    return associators(cc, cr, cop, assocClass, resultClass, role,
            resultRole, properties, 0);
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementAssociatorNames(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* resultClass,
    const char* role,
    const char* resultRole)
{
    return associators(cc, cr, cop, assocClass, resultClass, role,
            resultRole, NULL, 1);
}

static CMPIStatus references(
    const CMPIContext *ctx,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* role,
    const short names)
{
    CMPIStatus st;
    SwPackage sw_pkg;
    gchar *repo_id = NULL;
    CMPIObjectPath *op = NULL;
    char error_msg[BUFLEN] = "";

    st = lmi_class_path_is_a(_cb, KNameSpace(cop),
            LMI_SoftwareInstallationServiceAffectsElement_ClassName, assocClass);
    lmi_return_if_class_check_not_ok(st);

    init_sw_package(&sw_pkg);

    if (CMClassPathIsA(_cb, cop, LMI_SoftwareInstallationService_ClassName, &st)) {
        /* got SoftwareInstallationService - AffectingElement */
        if (role && strcasecmp(role, LMI_AFFECTING_ELEMENT) != 0) {
            lmi_debug("Unexpected role \"%s\".", role);
            goto done;
        }

        return enum_instances(ctx, cr, KNameSpace(cop), names);

    } else {
        /* got AffectedElement */
        if (role && strcasecmp(role, LMI_AFFECTED_ELEMENT) != 0)
            goto done;
        if (CMClassPathIsA(_cb, cop, LMI_SoftwareIdentity_ClassName, &st)) {
            /* got SoftwareIdentity */
            if (get_sw_pkg_from_sw_identity_op(cop, &sw_pkg) != 0) {
                goto done;
            }
            get_repo_id_from_sw_pkg(&sw_pkg, &repo_id, error_msg, BUFLEN);
            if (!repo_id) {
                goto done;
            }

        } else if (CMClassPathIsA(_cb, cop, CIM_ComputerSystem_ClassName, &st)) {
            /* got ComputerSystem */
            if (!lmi_check_computer_system(ctx, cop)) {
                KSetStatus(&st, ERR_NOT_FOUND);
                goto done;
            }
        } else {
            lmi_debug("Unknown association class \"%s\".",
                    CMGetCharsPtr(CMGetClassName(cop, &st), NULL));
            goto done;
        }

        st = get_software_installation_service_op(_cb, ctx, &op);
        if (st.rc != CMPI_RC_OK)
            goto done;

        LMI_SoftwareInstallationServiceAffectsElement w;
        LMI_SoftwareInstallationServiceAffectsElement_Init(&w, _cb, KNameSpace(cop));
        LMI_SoftwareInstallationServiceAffectsElement_SetObjectPath_AffectingElement(&w, op);
        LMI_SoftwareInstallationServiceAffectsElement_SetObjectPath_AffectedElement(&w, cop);

        if (names) {
            KReturnObjectPath(cr, w);
        } else {
            KReturnInstance(cr, w);
        }
    }

done:
    clean_sw_package(&sw_pkg);
    g_free(repo_id);

    if (*error_msg) {
        lmi_error("Failed to enumerate references: %s", error_msg);
        CMSetStatusWithChars(_cb, &st,
                st.rc == CMPI_RC_OK ? CMPI_RC_ERR_FAILED : st.rc, error_msg);
    }

    return st;
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementReferences(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* role,
    const char** properties)
{
    return references(cc, cr, cop, assocClass, role, 0);
}

static CMPIStatus LMI_SoftwareInstallationServiceAffectsElementReferenceNames(
    CMPIAssociationMI* mi,
    const CMPIContext* cc,
    const CMPIResult* cr,
    const CMPIObjectPath* cop,
    const char* assocClass,
    const char* role)
{
    return references(cc, cr, cop, assocClass, role, 1);
}

CMInstanceMIStub(
    LMI_SoftwareInstallationServiceAffectsElement,
    LMI_SoftwareInstallationServiceAffectsElement,
    _cb,
    LMI_SoftwareInstallationServiceAffectsElementInitialize(ctx))

CMAssociationMIStub(
    LMI_SoftwareInstallationServiceAffectsElement,
    LMI_SoftwareInstallationServiceAffectsElement,
    _cb,
    LMI_SoftwareInstallationServiceAffectsElementInitialize(ctx))

KONKRET_REGISTRATION(
    "root/cimv2",
    "LMI_SoftwareInstallationServiceAffectsElement",
    "LMI_SoftwareInstallationServiceAffectsElement",
    "instance association")
